/*
 *          Copyright (c) mjh-EDV Beratung, 1996-1999
 *     mjh-EDV Beratung - 63263 Neu-Isenburg - Rosenstrasse 12
 *          Tel +49 6102 328279 - Fax +49 6102 328278
 *                Email info@mjh.teddy-net.com
 *
 *   Author: Jordan Hrycaj <jordan@mjh.teddy-net.com>
 *
 *   $Id: peks-file.c,v 1.12 1999/03/31 17:59:54 renaud Exp $
 *
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Library General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Library General Public License for more details.
 *
 *   You should have received a copy of the GNU Library General Public
 *   License along with this library; if not, write to the Free
 *   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include "common-stuff.h"

#ifdef HAVE_CTYPE_H
# include <ctype.h>
#endif
#ifdef HAVE_PWD_H
# include <pwd.h>
#else
  struct passwd *getpwnam ();
#endif

#ifdef HAVE_SYS_STAT_H
# include <sys/stat.h>
#elif defined (HAVE_STAT_H)
# include <stat.h>
#else
  extern int stat ();
#endif
#ifndef HAVE_LSTAT
# define lstat(x,y) stat(x,y)
#endif

#if TIME_WITH_SYS_TIME
# include <sys/time.h>
# include <time.h>
#else
# if HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  include <time.h>
# endif
#endif

#include <fcntl.h>

#include "peks-internal.h"
#include "peks-file.h"

#ifdef USE_POLL_SLEEP
#ifdef HAVE_SYS_POLL_H
#include <sys/poll.h>
#elif defined (HAVE_POLL_H)
#include <poll.h>
#endif
#endif /* USE_POLL_SLEEP */

#include "peks-setup.h"
#include "peks-baseXX.h"
#include "peks-string.h"
#include "make-primes.h"
#include "messages.h"
#include "rnd-pool.h"
#include "crandom.h"
#include "hostnames.h"
#include "elgamal.h"
#include "baseXX.h"

/* the maximal line length in a key file */
#define MAX_LINE_LENGTH (5 * 1024)

/* ---------------------------------------------------------------------------- *
 *                        private helper functions                              *
 * ---------------------------------------------------------------------------- */

static int
private_access
 (const char *name,
  int     strictly)
{
   struct stat stbuf;

   if (lstat (name, &stbuf))
     return -1;
#ifdef HAVE_LSTAT
   /* check for regular file or symlink */
   if (S_ISLNK(stbuf.st_mode)) {
     errno = errnum (NO_SYMLINK_WANTED) ;
     return -1;
   }
#endif
   if (S_ISREG(stbuf.st_mode) == 0) {
     errno = errnum (REG_FILE_WANTED) ;
     return -1;
   }

   POINT_OF_RANDOM_STACK (21) ;

   if (stbuf.st_mode & (S_IWGRP|S_IWOTH)) {
     errno = errnum (NOT_GROUP_WRITABLE) ;
     return -1;
   }
   if (strictly && (stbuf.st_mode & S_IROTH)) {
     errno = errnum (NOT_WORLD_READABLE) ;
     return -1;
   }
   if (strictly > 1 && (stbuf.st_mode & S_IRGRP)) {
     errno = errnum (NOT_GROUP_READABLE) ;
     return -1;
   }

   POINT_OF_RANDOM (0,0);
   return 0;
}


static unsigned
valid_user_name
  (const char *s)
{
  if (s == 0
#ifdef USERNAME_NODGTSTART
      || isdigit (*s) /* must not start with a digit */
#endif
      )
    return 0 ;

  do {

    if (!isupper (*s) && 
	!islower (*s) &&
	!isdigit (*s) &&
	strchr (USERNAME_XCHARS, *s) == 0)
      return 0;
    
  } while (* ++ s != 0) ;

  return 1;
}



static unsigned
matches_line_type  /* evaluates numeric value of 2nd field entry */
  (const char *s,
   int      type)
{
  /* skip the header and advance to the first field */
  while (*s && !isspace (*s)) s++ ;
  while (*s &&  isspace (*s)) s++ ;

  /* advance to the second field */
  while (*s && !isspace (*s)) s++ ;
  while (*s &&  isspace (*s)) s++ ;
  
  if (*s == 0) return 0;

  /* expecting a single numeric letter - otherwise of type -1 */
  if (*s == 0 || s [1] != ' ')
    return type < 0 ;

  /* check, whether it matches the argument type */
  switch (s [0]) { /* is base 64 encoded */
  case 'A': return type == 0 ;
  case 'B': return type == 1 ;
  }

  return type == -1 ;
}


/*
 * test, whether a line 
 *
 *   <usr>@host: <field> <type-field> ...
 * or
 *   <usr>@ <field> <type-field> ...
 * or
 *   <host>: <field> <type-field> ...
 *
 * matches the selection criteria and return the position of <field>
 */


static const char *
start_of_key_fields
  (const char  *usr,
   const char *host,
   const char *line,
   int         type)
{
  const char *p;

  POINT_OF_RANDOM_VAR (p) ;
  
  /* both 0 are not allowed */
  if (usr == 0 && host == 0)
    return 0;
  
  /* skip leading spaces */
  while (*line && isspace (*line))
    ++ line ;

  POINT_OF_RANDOM_STACK (17) ;

  if (usr != 0) {
    unsigned len ;
    
    /* check for <user>@  ... */
    if ((p = strchr (line, '@')) == 0) 
      return 0 ;

    len = p - line;
    if (strncmp (usr, line, len) != 0 || usr [len] != 0)
      return 0;
    
    /* user matches, no host name given ? */
    if (host == 0)

      return /* stop at the pattern "<user>@ " */
	isspace (p [1]) && matches_line_type (p, type) 
	? p + 2 
	: 0 
	;
	
    /* user matches, check host name (host != 0) */
    line = p + 1;
  }

  POINT_OF_RANDOM_VAR (usr) ;

  /* find end of tag */
  if ((p = strchr (line + 1, ':')) == 0)
    return 0 ;

  return /* compare the host pattern "<tag>: " */
    isspace (p [1]) && matches_line_type (p, type) &&
    hostcmp (host, line, p - line) == 0
    ? p + 2 
    : 0 
    ;
}

/* sleep some milli secs */
static void
sleep_ms 
  (unsigned ms) 
{
  if (ms < 10)
    ms = 10 ;
  {
#ifdef USE_SELECT_SLEEP
    struct timeval tv;
    tv.tv_sec  =  0;
    tv.tv_usec = ms;
    select (FILE_LOCK_POLL_FD, 0, 0, 0, &tv);
#elif defined (USE_POLL_SLEEP)
    struct pollfd pfd;
    memset (&pfd, 0, sizeof (pfd)) ;
    pfd.fd = FILE_LOCK_POLL_FD;
    poll (&pfd, 1, ms) ;
#else
    sleep (1);
#endif
  }
}

static void
remove_stale_lockfile
  (char *file)
{
  struct stat stbuf;
  if (stat (file, &stbuf) == 0 &&
      stbuf.st_mtime + FILE_LOCK_STALE_SECS < time (0))
    /* remove stale file */
    unlink (file) ;
}

/* ---------------------------------------------------------------------------- *
 *                  raw public key file i/o functions                           *
 * ---------------------------------------------------------------------------- */

/*
 * parse a file linewise to find
 *
 *	<tag>: <key defs> ...
 *  or
 *	<user>@[<host>:] <key defs> ...
 *
 * and load that line into a key structure.
 */

peks_key *
read_peks_key
  (const char  *usr,
   const char  *tag,
   int       g_type, /* num val of 2nd entry, or -1 to disable */
   const char *file)
{
  FILE *inp;
  char read_buf [MAX_LINE_LENGTH] ;
  const char *s ;

  /* open input file if it exists */
  if ((usr == 0 && tag == 0) || file == 0) {
    errno = errnum (FUNCTION_ARG_ERROR) ;
    return 0;
  }

  /* check the usr name */
  if (usr != 0 && !valid_user_name (usr)) {
    errno = errnum (INVALID_USER_NAME) ;
    return 0 ;
  }

  POINT_OF_RANDOM_STACK (13) ;

  if ((inp = fopen (file, "r")) == 0) {
    char *new ;
    int n, save = errno ;
    time_t limit; 

    /* create output file name for "new" */
    new = ALLOCA (strlen (file) + 5) ;
    strcat (strcpy (new, file), ".new");

    /* check, whether this file is being edited */
    if ((n = open (new, O_RDONLY)) < 0) {
      if (save == ENOENT)
	errno = 0 ;
      return 0 ;
    }

    remove_stale_lockfile (new) ;
  
    /* wait and retry untill the file is ready from editing*/
    limit = time (0) + FILE_LOCK_MAXW_SECS ;
    do { 
      close (n) ;
      fast_random_bytes ((char*)&n, sizeof (n));
      sleep_ms (n % FILE_LOCK_WAITMOD_MS+1);
    } while ((inp = fopen (file, "r")) == 0 &&
	     time (0) <= limit && 
	     (n = open (new, O_RDONLY)) >= 0) ;

    if (inp == 0 && (inp = fopen (file, "r")) == 0) {
      errno = errnum (FILE_IS_BUSY) ;
      return 0;
    }
  }

  if (private_access (file, 2) < 0) {
    fclose (inp);
    return 0;
  }

  /* parse input file up until the tag line */
  while (fgets (read_buf, sizeof (read_buf), inp)) {

    if (read_buf [0] == '#')
      continue ;
    
    if ((s = start_of_key_fields (usr, tag, read_buf, g_type)) != 0) {
      fclose (inp) ;
      return peks_parse_encryption_line (s) ;
    }
  }
    
  /* did not find that tag */
  errno = 0 ;
  fclose (inp);
  return 0;
}



int
save_peks_key
(peks_key    *key, /* key == 0: delete entry */
   const char  *usr,
   const char  *tag,
   int       g_type, /* num val of 2nd entry, or -1 to disable */
   const char *file)
{
  FILE *inp = 0, *outp = 0;
  char read_buf [MAX_LINE_LENGTH], *new, *old, *s ;
  unsigned len, r_code = -1 ;
  int n, omask, newfd ;

  if ((usr == 0 && tag == 0) || file == 0) {
    errno = errnum (FUNCTION_ARG_ERROR) ;
    return -1 ;
  }

  /* check the usr name */
  if (usr != 0 && !valid_user_name (usr)) {
    errno = errnum (INVALID_USER_NAME) ;
    return 0 ;
  }

  POINT_OF_RANDOM_VAR (key);

  /* set private file creation mask */
  omask = umask (07177) ;

  POINT_OF_RANDOM_VAR (inp) ;

  /* create output file names for "new" and "old" */
  new = ALLOCA ((len = strlen (file)) + 5) ;
  strcat (strcpy (new, file), ".new") ;

  POINT_OF_RANDOM_STACK (27) ;

  old = ALLOCA (len + 5) ;
  strcat (strcpy (old, file), ".old");

  POINT_OF_RANDOM_VAR (new) ;

  /* create output file */
  if ((newfd = open (new, O_WRONLY|O_EXCL|O_CREAT, 0600)) < 0) 
    remove_stale_lockfile (new) ;
  
  /* retry if we failed */
  if (newfd < 0 &&
      (newfd = open (new, O_WRONLY|O_EXCL|O_CREAT, 0600)) < 0) {

    int repeat = 1;
    do { /* repeat once */
      
      time_t limit = time (0) + FILE_LOCK_MAXW_SECS ;

    /* Somebody else is editing that file ?  So wait and retry
       untill the lock file (aka new) has vanished */
    
      do {
	fast_random_bytes ((char*)&n, sizeof (n));
	sleep_ms (n % FILE_LOCK_WAITMOD_MS+1);
      } while (time (0) <= limit && 
	       (newfd = open (new, O_WRONLY|O_EXCL|O_CREAT)) < 0);
      
      if (newfd >= 0) 
	break ;

      /* just double check everything */
      if ((n = open (file, O_RDONLY)) >= 0) {
	close (n) ;
	errno = errnum (FILE_IS_BUSY) ;
	goto get_ready_for_return;
      }

      /* so the file <new> is at least FILE_LOCK_MAXWAIT_MS msecs old, and the 
	 file <file> does not exist - and we conclude that <new> is stale */
      unlink (new) ;
      
    } while (-- repeat >= 0) ;

    if (newfd < 0 && (newfd = open (new, O_WRONLY|O_EXCL|O_CREAT)) < 0) {
      errno = errnum (FILE_IS_BUSY) ;
      goto get_ready_for_return;
    }
  }
  if ((outp = fdopen (newfd, "w")) == 0) {
    errno = errnum (CANT_OPEN_KEY_WFILE) ;
    goto get_ready_for_return;
  }

  /* open input file if it exists */
  if ((inp = fopen (file, "r")) == 0 && errno != ENOENT) {
    errno = errnum (CANT_OPEN_KEY_RFILE) ;
    goto get_ready_for_return;
  }
  if (inp != 0) {

    /* check file permissions */
    if (private_access (file, 2) < 0) 
      goto get_ready_for_return;
  
    /* parse input file up until the tag line */
    while (fgets (read_buf, sizeof (read_buf), inp) > 0) {
      
      if (read_buf [0] != '#') {

	if (start_of_key_fields (usr, tag, read_buf, g_type) != 0)
	  break ;
      
	/* check for input line overflow */
	n = strlen (read_buf) ;
	if (read_buf [n-1] != '\n') {
	  errno = errnum (KEYFILE_PARSER_ERR) ;
	  goto get_ready_for_return;
	}
      }
      
      /* copy to the output file */
      fputs (read_buf, outp) ;
    }
  } /* if (inp != 0) */ 
  
  POINT_OF_RANDOM_VAR (old) ;

  if (key != 0) {
    /* make string from private key, make whole keys line */
    s = make_peks_key_line (usr, tag, key) ;
    fputs (s, outp);
    fputc ('\n', outp) ;

    POINT_OF_RANDOM_VAR (s) ;
    /* clean up after writing the keys string to file */
    XFREE (s);
  }

  if (inp != 0) {

    /* dump the rest onto file */
    while (fgets (read_buf, sizeof (read_buf), inp) > 0) {
      
      /* check for input line overflow */
      n = strlen (read_buf) ;
      if (read_buf [n-1] != '\n' &&
	  read_buf [n-1] != '\r') {
	errno = errnum (KEYFILE_PARSER_ERR) ;
	goto get_ready_for_return;
      }

      /* copy to the output file */
      fputs (read_buf, outp) ;
    }
  
    fclose (inp);
  } /* if (inp != 0) */ 

  fclose (outp);

  /* files have been closed, now */
  inp = outp = 0 ;

  /* save current version of the key file */
  if (inp != 0) {
    unlink (old) ;
    rename (file, old) ;
  }

  /* install new key file */
  r_code = rename (new, file) ;

 get_ready_for_return:

  if (inp != 0)  fclose (inp);
  if (outp != 0) fclose (outp);
  DEALLOCA (new) ;
  DEALLOCA (old) ;
  umask (omask) ; /* restore file creation mode */
  return r_code ;
}
/* ---------------------------------------------------------------------------- *
 *               private passwd file management                                 *
 * ---------------------------------------------------------------------------- */

static int
peks_write_passwd 
  (const peks_key  *k,
   const char    *usr,
   const char   *host,
   char         count,
   const char *passwd,
   const char   *file)
{
  peks_key *user_pwd ;

  mpz_t PUB ;
  char *s ;
  int n ;

  if (usr == 0 || file == 0) {
    errno = errnum (FUNCTION_ARG_ERROR) ;
    return -1 ;
  }

  POINT_OF_RANDOM_STACK (5);

  if (passwd == 0 || k == 0)  /* delete */
    return save_peks_key (0, usr, host, 0, file) ;

  user_pwd = memset (ALLOCA (sizeof (peks_key)), 0, sizeof (peks_key)) ;

  mpz_init (user_pwd->modulus) ;
  mpz_init (user_pwd->private) ;
  
  mpz_init_set_ui (PUB, k->generator) ;
  mpz_powm (PUB, PUB, k->private, k->modulus);	/* pub key := g^a (mod p) */

  n = 2 + strlen (usr) + 1 + strlen (passwd) ;
  sprintf (s = ALLOCA (n+1), "%c %s %s", ('0' + count) & 0x7f, usr, passwd) ;
  
  n = el_gamal_encrypt 
    (&user_pwd->modulus, &user_pwd->private, &k->modulus, k->generator, &PUB, s, n) ;
  
  POINT_OF_RANDOM_VAR (s);
 
  DEALLOCA (s) ;
  mpz_clear (PUB) ;

  if (n >= 0) /* user_pwd == 0 <=> delete */
    n = save_peks_key (user_pwd, usr, host, 0, file) ;

  mpz_clear (user_pwd->modulus) ;
  mpz_clear (user_pwd->private) ;
  DEALLOCA  (user_pwd) ;

  return n;
}


static char *
peks_read_passwd 
  (const peks_key *k,
   const char   *usr,
   const char  *host,
   const char  *file)
{
  peks_key *user_pwd ;
  char *pwd, *s ;
  unsigned n;

  if (k == 0 || usr == 0 || file == 0) {
    errno = errnum (FUNCTION_ARG_ERROR) ;
    return 0 ;
  }

  if ((user_pwd = read_peks_key (usr, host, 0, file)) == 0) {
    if (errno == 0)
      errno = errnum (NO_SUCH_USER) ;
    return 0 ;
  }
  
  pwd = b64_el_gamal_decrypt
    (&user_pwd->modulus, &user_pwd->private, &k->modulus, &k->private) ;

  end_peks_key (user_pwd) ;

  if (pwd == 0)
    return 0;

  s = base64toBin (pwd, &n) ;
  XFREE (pwd) ;

  return s;
}


/* ---------------------------------------------------------------------------- *
 *               public key file functions                                      *
 * ---------------------------------------------------------------------------- */

int
check_peks_sender_key
  (peks_key    *key,
   const char  *usr,
   const char *host,
   unsigned  create,
   const char *file)
{
  char *p = 0;
  int r_code = -1 ;

  peks_key *fkey ;

  /* initialize prime random generator (unless done, yet) */
  init_random_gen ((char*)&fkey, sizeof (fkey)) ;

  if (key == 0 || file == 0) {
    errno = errnum (FUNCTION_ARG_ERROR) ;
    return -1;
  }

  POINT_OF_RANDOM_VAR (file) ;

  /* prepend by HOME variable */
  if (file [0] == '~' && file [1] == '/') {
    char *s = getenv ("HOME") ;
    if (s == 0) {
      errno = errnum (CANT_EXPAND_FNAME) ;
      return -1;
    }
    POINT_OF_RANDOM_VAR (p) ;
    p = strcpy (ALLOCA (strlen (s) + strlen (file)), s) ;
    file = strcat (p, file + 1) ;
  }

  POINT_OF_RANDOM_VAR (file) ;

  /* check, whether the key file exists, already */
  if ((fkey = read_peks_key (usr, host, -1, file)) == 0 && errno)
    goto done ;
  /* re-check for a user entry without host spec */
  if (fkey == 0 && usr != 0)
    if ((fkey = read_peks_key (usr, 0, -1, file)) == 0 && errno)
      goto done ;

  POINT_OF_RANDOM_VAR (fkey) ;
  
  if (fkey != 0) {
    /* check the key in the file against the key, given */
    if (mpz_cmp (key->modulus, fkey->modulus) == 0 &&
	mpz_cmp (key->private, fkey->private) == 0 &&
	key->generator == key->generator)
      r_code = 0 ;
    else
      errno = errnum (SPOOF_ALERT) ;
    end_peks_key (fkey) ;
    goto done ;
  }

  /* ok, set the not-found flag */
  r_code = 1 ;

  if (create) {
    /* append the key in the key file */
    if (save_peks_key (key, 0, host, -1, file) >= 0) {
      errno = errnum (STORED_SERVER_KEY) ;
    } else
      r_code = -1 ;
  }
  
 done:

  if (p != 0)
    DEALLOCA (p);
 
  return r_code ;
}


peks_key *
read_peks_keyfile
  (const char *usr,
   const char *host,
   const char *file)
{
  peks_key *key ;
  char *p = 0;

  errno = 0;

  POINT_OF_RANDOM_STACK (13) ;
  
  /* get user from password structure */
  if (usr == 0) {
    struct passwd *pwd = getpwuid (getuid ()) ;
    if (pwd == 0 || (usr = pwd->pw_name) == 0) {
      if (errno == 0)
	errno = errnum (NO_USER_NAME) ;
      return 0 ;
    }
  }
  /* get host name */
  if (host == 0 && (host = get_my_host_name ()) == 0) {
    if (errno == 0)
      errno = errnum (NO_HOST_NAME) ;
    return 0 ;
  }
  /* prepend by HOME variable */
  if (file [0] == '~' && file [1] == '/') {
    char *s = getenv ("HOME") ;
    if (s == 0) {
      errno = errnum (CANT_EXPAND_FNAME) ;
      return 0;
    }
    POINT_OF_RANDOM_VAR (p) ;
    p = strcpy (ALLOCA (strlen (s) + strlen (file)), s) ;
    file = strcat (p, file + 1) ;
  }

  /* initialize prime random generator (unless done, yet) */
  init_random_gen ((char*)&key, sizeof (key)) ;

  /* search for "<usr>@<host>: " first and "<usr>@ ", next */
  if ((key = read_peks_key (usr, host, -1, file)) == 0)
    key = read_peks_key (usr, 0, -1, file) ;

  if (p != 0)
    DEALLOCA (p);
  
  POINT_OF_RANDOM_VAR (key) ;

  /* some non key entries have 0 or 1 generators */
  if (key == 0 || key->generator > 1)
    return key ;

  POINT_OF_RANDOM_VAR (file) ;

  /* so this is never a private key */
  errno = errnum (MODIFIED_KEY_TAG) ;
  end_peks_key (key);
  return 0;
}


int
create_peks_keyfile
  (peks_key    *key,
   const char *usr,
   const char *host,
   const char *file)
{
  int r_code ;
  char *p = 0;

  errno = 0;

  POINT_OF_RANDOM_STACK (13) ;
  
  /* get user from password structure */
  if (usr == 0) {
    struct passwd *pwd = getpwuid (getuid ()) ;
    if (pwd == 0 || (usr = pwd->pw_name) == 0) {
      if (errno == 0)
	errno = errnum (NO_USER_NAME) ;
      return -1 ;
    }
  }
  /* get host name */
  if (host == 0 && (host = get_my_host_name ()) == 0) {
    if (errno == 0)
      errno = errnum (NO_HOST_NAME) ;
    return -1 ;
  }
  /* prepend by HOME variable */
  if (file [0] == '~' && file [1] == '/') {
    char *s = getenv ("HOME") ;
    if (s == 0) {
      errno = errnum (CANT_EXPAND_FNAME) ;
      return -1;
    }
    POINT_OF_RANDOM_VAR (p) ;
    p = strcpy (ALLOCA (strlen (s) + strlen (file)), s) ;
    file = strcat (p, file + 1) ;
  }

  POINT_OF_RANDOM_VAR (file) ;

  /* search for "<usr>@<host>: " first and "<usr>@ ", next */
  if ((r_code = save_peks_key (key, usr, host, -1, file)) < 0)
    r_code = save_peks_key (key, usr, 0, -1, file) ;

  if (p != 0)
    DEALLOCA (p);

  return r_code ;
}


/* ---------------------------------------------------------------------------- *
 *               public key convenience function                                *
 * ---------------------------------------------------------------------------- */

peks_key *
peks_private_key
  (const char *uath,
   const char *file,
   unsigned  keylen)
{
  peks_key *key ;
  char msg [300] ;
  unsigned n ;

  char *usr, *host;
  
  if (uath == 0) 
    uath = "";

  /* must be used in the current { frame } */
  usr  = ALLOCASTRDUP (uath) ;
  host = strchr (usr, '@') ;
  
  if (host != 0) {
    * host = '\0' ;
    host ++ ;
  } else
    if (*uath == '\0') {
      DEALLOCA (usr) ;
      usr = 0 ;
    }

  POINT_OF_RANDOM_STACK (13) ;

  if (keylen < MINIMUM_PRIME_BITS)
    keylen = MINIMUM_PRIME_BITS ;

  if ((key = read_peks_keyfile (usr, host, file)) == 0) {
    if (errno) {
      if (file == 0)
	file = "(*null*)" ;
      else if (strlen (file) > 100)
	file = "<very-long-path>" ;
      sprintf (msg, peks_strerr (errnum (CANT_READ_KEYS_SS)),
	       file, peks_strerr (errno));
      errno = peks_errnum (msg) ;
      DEALLOCA (usr) ;
      return 0 ;
    }
    POINT_OF_RANDOM_VAR (key) ;
    if ((key = new_peks_key (keylen)) == 0) {
      sprintf (msg, peks_strerr (errnum (CANT_GEN_KEYS_S)),
	       peks_strerr (errno));
      errno = peks_errnum (msg) ;
      DEALLOCA (usr) ;
      return 0 ;
    }
    if (create_peks_keyfile (key, usr, host, file) < 0) {
      if (strlen (file) > 100)
	file = "<very-long-path>" ;
      sprintf (msg, peks_strerr (errnum (CANT_SAVE_KEYS_SS)),
	       file, peks_strerr (errno));
      errno = peks_errnum (msg) ;
      DEALLOCA (usr) ;
      return 0 ;
    }
  }

  POINT_OF_RANDOM_STACK (5);

  if ((n = peks_keylen (key)) < keylen) {
    end_peks_key (key);
    sprintf (msg, peks_strerr (errnum (PRVKEY_2SHORT_UU)), n, keylen);
    errno = peks_errnum (msg) ;
    DEALLOCA (usr) ;
    return 0 ;
  }

  /* make a basxe 64 public key */
  key->pubkey_str = make_pubkey_line (key) ;

  POINT_OF_RANDOM_VAR (file) ;

  DEALLOCA (usr) ;
  return key ;
}


char *
peks_edit_passwd_file
  (const peks_key *k,
   const char  *uath,
   unsigned    count,
   const char   *pwd,
   const char  *file)
{
  char *host, *usr, *s ;

  if (uath == 0) {
    errno = errnum (NO_USER_NAME) ;
    return 0;
  }

  usr  = ALLOCASTRDUP (uath) ;
  host = strchr (usr, '@') ;
  
  if (host != 0) {
    * host = '\0' ;
    host ++ ;
  }

  if ((k == 0 && pwd == 0) ||   /* delete request */
      (k != 0 && pwd != 0))     /* edit request */
    s = peks_write_passwd (k, usr, host, count, pwd, file) < 0 ? 0 : "" ;
  else
    /* view request */
    s = peks_read_passwd (k, usr, host, file) ;

  DEALLOCA (usr) ;
  return s ;
}


int
peks_delete_userkey
  (const char  *uath,
   const char  *file)
{
  char *usr, *host ;
  peks_key *k ;
  int n;
  
  if (uath == 0) {
    errno = errnum (NO_USER_NAME) ;
    return -1;
  }

  usr  = ALLOCASTRDUP (uath) ;
  host = strchr (usr, '@') ;
  
  if (host != 0) {
    * host = '\0' ;
    host ++ ;
  }

  if ((k = read_peks_key (usr, host, -1, file)) == 0) {
    n = -1 ;
    errno = errnum (NO_SUCH_USER); 
  } else {
    end_peks_key (k) ;
    n = save_peks_key (0, usr, host, -1, file) ;
  }
  
  DEALLOCA (usr) ;
  return n ;
}

int
peks_list_keyfile
  (void (*prt) (const char*),
   const char          *file)
{
  FILE *inp;
  char read_buf [MAX_LINE_LENGTH] ;
  char  prt_buf [MAX_LINE_LENGTH] ;
  char *s ;

  /* open input file if it exists */
  if (private_access (file, 2) < 0) {
    if (errno == ENOENT)
      errno = 0 ;
    return 0;
  }

  if (prt == 0)
    prt = xprint1 ;
  
  POINT_OF_RANDOM_STACK (13) ;

  if ((inp = fopen (file, "r")) == 0)
    return 0 ;

  /* parse input file up until the tag line */
  while (fgets (read_buf, sizeof (read_buf), inp)) {

    char *key_type, *key_code ;
    
    if (read_buf [0] == '#')
      continue ;

    /* check the tag */
    if (((s = strchr (read_buf, ':')) == 0 &&
	 (s = strchr (read_buf, '@')) == 0) ||
	!isspace (s [1]))
      continue ;
    s [1] = '\0' ;

    /* get the tag */
    key_type = (strchr (read_buf, '@') != 0 ? "user" : "host") ;
    s [0] = '\0' ;

    /* skip leading spaces */
    s += 2 ;
    while (*s && isspace (*s))
      ++ s ;

    /* skip the module field */
    while (*s && !isspace (*s))
      ++ s ;

    /* skip leading spaces */
    while (*s && isspace (*s))
      ++ s ;

    if (* s == '\0')
      continue ;

    key_code = (s [0] == 'A' && isspace (s [1]) ? "password" : "key") ;

    sprintf (prt_buf, "%20s - %s %s\n", read_buf, key_type, key_code);
    (*prt) (prt_buf) ;
  }
    
  fclose (inp);
  return 0;
}
