/*
 *          Copyright (c) mjh-EDV Beratung, 1996-1999
 *     mjh-EDV Beratung - 63263 Neu-Isenburg - Rosenstrasse 12
 *          Tel +49 6102 328279 - Fax +49 6102 328278
 *                Email info@mjh.teddy-net.com
 *
 *   Author: Jordan Hrycaj <jordan@mjh.teddy-net.com>
 *
 *   $Id: peks-rpc.c,v 1.2 1999/03/31 17:59:56 renaud Exp $
 *
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Library General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Library General Public License for more details.
 *
 *   You should have received a copy of the GNU Library General Public
 *   License along with this library; if not, write to the Free
 *   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "common-stuff.h"

#ifdef HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif

#ifdef HAVE_CTYPE_H
# include <ctype.h>
#endif

#include "peks-internal.h"
#include "peks-rpc.h"
#include "peks-string.h"
#include "rnd-pool.h"
#include "iostream.h"
#include "messages.h"
#include "version.h"


/* ---------------------------------------------------------------------------- *
 *                  private data and structures                                 *
 * ---------------------------------------------------------------------------- */

typedef struct _cb_data {

  int myself, fd ;

  prpc          *parent;
  struct _cb_data *next;

} cb_data ;


static cb_data *cb_table ;

/* ---------------------------------------------------------------------------- *
 *                     private helpers                                          *
 * ---------------------------------------------------------------------------- */


/* ---------------------------------------------------------------------------- *
 *                     private functions                                        *
 * ---------------------------------------------------------------------------- */


/* ---------------------------------------------------------------------------- *
 *                private call back functions                                   *
 * ---------------------------------------------------------------------------- */
static int 
initial_tcatcher 
  (char      *buf, 
   unsigned   len, 
   int     socket, 
   void **context) 
{
  cb_data *t, *ctx = *context;
  
  if (buf == 0 && ctx != 0) {

    /* unlink from call back data table */
    if ((t = cb_table) == ctx)
      cb_table = ctx->next ;
    else
      while (t != 0) {
	if (t->next == ctx) {
	  t->next = ctx->next ;
	  break ;
	}
	t = t->next ;
      }

    /* unlink parent */
    if (ctx->parent != 0) {
      ctx->parent->tdata = 0;
      ctx->parent->fd = -1 ;
      ctx->parent->tid = 0 ;
    }

    XFREE (ctx);
    return 0 ;
  }

  if (ctx == 0) {
    ctx = XMALLOC (sizeof (*ctx)) ;

    /* link to call back data table */
    ctx->next = cb_table ;
    cb_table  = ctx ;
      
    /* activate as context */
    *context = ctx ;

    /* get my thread id */
    ctx->myself = io_ctrl (socket, IO_CATCH_THREAD, 0, 0) ;
    ctx->fd     = socket ;
  }

  return len;
}      


/* ---------------------------------------------------------------------------- *
 *                      public rpc server functions                             *
 * ---------------------------------------------------------------------------- */

prpc *
prpc_accept
  (int    fd,
   prpc *rpc)
{
  /* this is the ordered list of authentication schemes, available */
  const char *proto [] = {"prpc", 0};

  int (*fn) (char*,unsigned,int,void**) = initial_tcatcher ;

  char *s, *buf;
  int version, n, stop_state;
  cb_data *t; 

  if (rpc != 0) {
    if (rpc->fd == fd && rpc->tid >= 0) {
      errno = errnum (PRPC_ALREADY_INIT) ;
      return 0 ;
    }
  } else
    rpc = XMALLOC (sizeof (prpc)) ;

  buf = ALLOCA (1025) ;
  
  /* we do not need to see administrative 0-packets */
  stop_state = io_ctrl (fd, IO_STOPONEMPTY_STATE, (n=0, &n), 0) ;

  POINT_OF_RANDOM_VAR (buf);

  /* install the thread catcher */
  if ((rpc->tid = io_ctrl (fd, IO_CATCH_THREAD, &fn, 0)) < 0)
    goto return_error;

  POINT_OF_RANDOM_VAR (fn);

  /* get the open request from the client */
  if ((n = io_recv (fd, buf, 1024, 0)) < 0)
    goto return_error;
  buf [n] = '\0' ;

  /* link to call back function data */
  t = cb_table;
  while (t != 0 && t->fd != fd && t->myself != rpc->tid) 
    t = t->next ;
  if (t == 0) { /* cannot happen :-) */
    errno = errnum (PRPC_INIT_FAILED);
    goto return_error ;
  }
  rpc->tdata = t ;		/* pointer to catcher data aerea */
  t->parent = rpc ;		/* reverse pointer */

  POINT_OF_RANDOM_VAR (rpc);
  
  /* parse the protocol version */
  if ((s = strchr (buf, ':')) == 0 || 
      !(++s, isspace (*s)) ||
      (version = peks_split_ident (proto, buf, s - buf)) == 0) {
    errno = errnum (PRPC_HEADER_FAILED) ;
    goto return_error ;
  }
  
  /* only "prpc/x.x" supported, here */
  if (version < 10000 || 20000 <= version) {
    errno = errnum (PRPC_HEADER_FAILED) ;
    goto return_error ;
  }
  /* no compat mode, necessary, yet */
  if (version != (100 + PEKSMAJOR) * 100 + PEKSMINOR) {
    errno = errnum (PRPC_HEADER_FAILED) ;
    goto return_error ;
  }

  /* must be a connect request */
  if (strncmp (++s, "connect", 7) != 0) {
    errno = errnum (PRPC_COMMD_ILLEGAL) ;
    goto return_error ;
  }

  /* accept it */
  if (io_send (fd, "ok", 3, 0) < 0)
    goto return_error;

  DEALLOCA (buf);
  io_ctrl (fd, IO_STOPONEMPTY_STATE, &stop_state, 0) ;
  return rpc;

 return_error:
  
  n = errno ;
  io_ctrl (fd, IO_STOPONEMPTY_STATE, &stop_state, 0) ;
  io_ctrl (fd, IO_UNCATCH_THREAD, &rpc->tid, 0);
  io_ctrl (fd, IO_UNCATCH_THREAD, 0, 0);
  DEALLOCA (buf);
  XFREE    (rpc);
  return 0;
}


/* ---------------------------------------------------------------------------- *
 *                      public rpc client functions                             *
 * ---------------------------------------------------------------------------- */

#if 0
int
prpc_dispatch
  (prpc *rpc,
   char *svc)
{
  int n, payload = 30 + strlen (svc) + rpc->size;
  char *p, *buf;

  if (payload < 1024)
    payload = 1024 ;

  buf = ALLOCA (payload+1) ;

  /* assemble transfer record: "req: <svc> <size>: <binary data ...> */
  sprintf (p = buf, "req: %s %u: ", svc, rpc->size);
  p += strlen (buf) ;
  if (rpc->size != 0)
    memcpy (p, rpc->data, rpc->size) ;
  p += rpc->size ;

  /* send on the rpc channel */
  if (io_send (rpc->fd, buf, p - buf, 0) < 0)
    goto return_error;

  POINT_OF_RANDOM_VAR (svc);
  
  /* get answer */
  if ((n = io_recv (rpc->fd, buf, payload, 0)) < 0)
    goto return_error;
  buf [n] = '\0';

  POINT_OF_RANDOM_VAR (buf);

  return 0 ;
  
 return_error:
  
  n = errno ;
  DEALLOCA (buf);
  errno = n ;
  return -1 ;
}
#endif


prpc *
prpc_connect
  (int    fd,
   prpc *rpc)
{
  int pid, n, stop_state ;
  char *buf;

  if (rpc != 0) {
    if (rpc->fd == fd && rpc->tid >= 0) {
      errno = errnum (PRPC_ALREADY_INIT) ;
      return 0 ;
    }
  } else
    rpc = XMALLOC (sizeof (prpc)) ;

  /* create control channel and activate it */
  if ((rpc->tid = io_ctrl (fd, IO_REGISTER_THREAD, 0, 1)) < 0)
    return 0;

  /* we do not need to see administrative 0-packets */
  stop_state = io_ctrl (fd, IO_STOPONEMPTY_STATE, (n=0, &n), 0) ;

  POINT_OF_RANDOM_VAR (rpc);

  buf = ALLOCA (1025) ;
  if ((pid = io_ctrl (fd, IO_ACTIVATE_THREAD, &rpc->tid, 1)) < 0 ) 
    goto return_error;

  /* save socket id */
  rpc->fd = fd;

  /* negotiate rpc channel */
  sprintf (buf, "%s: connect", PRPC_VERSION);
  if (io_send (fd, buf, strlen (buf)+1, 0) < 0)
    goto return_error;

  POINT_OF_RANDOM_VAR (fd);
  
  /* get answer */
  if ((n = io_recv (fd, buf, 1024, 0)) < 0)
    goto return_error;
  buf [n] = '\0';

  if (strncmp (buf, "ok", 2) != 0)
    goto return_error;

  POINT_OF_RANDOM_VAR (buf);
  
  /* restore previously active thread id */
  if (io_ctrl (fd, IO_ACTIVATE_THREAD, &pid, 1) >= 0) {
    DEALLOCA (buf);
    io_ctrl (fd, IO_STOPONEMPTY_STATE, &stop_state, 0) ;
    return rpc;
  }
  
 return_error:
  
  n = errno ;
  io_ctrl (fd, IO_DESTROY_THREAD, &rpc->tid, 1);
  io_ctrl (fd, IO_STOPONEMPTY_STATE, &stop_state, 0) ;
  XFREE    (rpc);
  DEALLOCA (buf);
  errno = n ;
  return 0 ;
}
