/*
 *          Copyright (c) mjh-EDV Beratung, 1996-1999
 *     mjh-EDV Beratung - 63263 Neu-Isenburg - Rosenstrasse 12
 *          Tel +49 6102 328279 - Fax +49 6102 328278
 *                Email info@mjh.teddy-net.com
 *
 *   Author: Jordan Hrycaj <jordan@mjh.teddy-net.com>
 *
 *   $Id: peks-svc.c,v 1.2 1999/03/31 17:59:57 renaud Exp $
 *
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Library General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Library General Public License for more details.
 *
 *   You should have received a copy of the GNU Library General Public
 *   License along with this library; if not, write to the Free
 *   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *   PSVC - peks supervisor call for rpc
 */

#include "common-stuff.h"

#ifdef HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif

#ifdef HAVE_CTYPE_H
# include <ctype.h>
#endif

#include "peks-internal.h"
#include "peks-svc.h"
#include "peks-string.h"
#include "rnd-pool.h"
#include "iostream.h"
#include "messages.h"
#include "version.h"

/* ---------------------------------------------------------------------------- *
 *                  private data and structures                                 *
 * ---------------------------------------------------------------------------- */

typedef struct _svc_proto {

  const char           *fmt ;	/* function argument list prototype */
  int                 laxen ;	/* allow missing parameter following */
  void                 *env ;	/* environment pointer for that function */
  int (*exec)(void*,void**) ;	/* the rpc function to be executed */
  struct _svc_proto *next;	/* next prototype for the same function */

} svc_proto ;


typedef struct _garbage {

  struct _garbage * next ;
  char *data [1] ;
  /*VARIABLE SIZE*/

} garbage ;


/* ---------------------------------------------------------------------------- *
 *                     private helpers                                          *
 * ---------------------------------------------------------------------------- */

#ifndef XREALLOC
static void *
xrealloc (void *p, unsigned n)
{
  void *q = realloc (p, n) ;

  if (q == 0) {
    fprintf (stderr, "Could not re-allocate to %u bytes !\n", n);
    exit (1);
  }
  
  return q ;
}

#define XREALLOC_DOES_NOT_INITALIZE
#define XREALLOC(p,n)	xrealloc (p, n)
#endif


static void *
yalloc 
  (prpc_data *rpc,
   unsigned     n)
{
  garbage *g = XMALLOC (sizeof (garbage) - 1 + n);

  /* record data allocated */
  g->next = rpc->garbage ;
  rpc->garbage = g ;

  return g->data ;
}


static void
yfree
  (prpc_data *rpc,
   garbage  *stop)
{
  garbage *g ;

  /* free the first part up until stop */
  while ((g = rpc->garbage) != 0 && g != stop) {
    rpc->garbage = g->next ;
    XFREE (g);
  }
}



#define buffer2ulong(b)    ( (((b) [0] << 24) & 0xff000000) |\
			     (((b) [1] << 16) & 0x00ff0000) |\
			     (((b) [2] <<  8) & 0x0000ff00) |\
		 	     (((b) [3]      ) & 0x000000ff) )

#define  ulong2buffer(b,u)  ( (b) [0] = ((u) >> 24) & 0xff ,\
                              (b) [1] = ((u) >> 16) & 0xff ,\
                              (b) [2] = ((u) >>  8) & 0xff ,\
                              (b) [3] = ((u)      ) & 0xff )

#define buffer2ushort(b)   ( (((b) [0] << 8) & 0xff00) |\
			     (((b) [1]     ) & 0x00ff) )

#define ushort2buffer(b,u) ( (b) [0] = ((u) >> 8) & 0xff ,\
                             (b) [1] = ((u)     ) & 0xff )

#define bufferXbuffer2ulong(b1, b2) ( ((((b1) [0] ^ (b2) [0]) << 24) & 0xff000000) |\
				      ((((b1) [1] ^ (b2) [1]) << 16) & 0x00ff0000) |\
				      ((((b1) [2] ^ (b2) [2]) <<  8) & 0x0000ff00) |\
				      ((((b1) [3] ^ (b2) [3])      ) & 0x000000ff) )

#define ulongXbuffer2buffer(t, u, b) ( (t) [0] = (((u) >> 24) ^ (b) [0]) & 0xff ,\
				       (t) [1] = (((u) >> 16) ^ (b) [1]) & 0xff ,\
				       (t) [2] = (((u) >>  8) ^ (b) [2]) & 0xff ,\
				       (t) [3] = (((u)      ) ^ (b) [3]) & 0xff )

/* ---------------------------------------------------------------------------- *
 *                     private functions                                        *
 * ---------------------------------------------------------------------------- */

static char *
append2rpc_spooler_prep
  (prpc_data *rpc,
   unsigned   mul,
   char      *idf,
   unsigned  size)
{
  char *p, buf [20] ;

  if (rpc->size + size > MAX_FMT_MULTIPLIER) {
    errno = errnum (PSVC_SIZE_2LARGE) ;
    return 0 ;
  }

  /* make a format string: [<multiplier>]<identifier> */
  switch (mul) {
  case 1: 
    strcpy (buf, idf) ;
    break;
  default: 
    sprintf (buf, "%u%s", mul, idf) ;
  }

  /* extend the format string */
  if (rpc->fmt == 0) {
    rpc->fmt = XSTRDUP (buf) ;
  } else {
    rpc->fmt = XREALLOC (rpc->fmt, strlen (rpc->fmt) + strlen (buf) + 1) ;
    strcat (rpc->fmt, buf);
  }

  /* extend the transfer buffer */
  if (rpc->data == 0) {
    rpc->data = XMALLOC (size) ;
    rpc->size = 0 ;
  } else 
    rpc->data = XREALLOC (rpc->data, rpc->size + size) ;

  /* return where you can append data */
  p = rpc->data + rpc->size ;

  /* update to new size */
  rpc->size += size;

  return p;
}


static unsigned
valid_psvc_name
  (const char *p)
{
  if (p == 0) {
    errno = errnum (PSVC_INT_ARG_ERROR);
    return 0;
  }

#ifdef PSVCNAME_NODGTSTART
  if (isdigit (*p)) { 
    /* must not start with a digit */
    errno = errnum (PSVC_ILLCHAR_IN_SVC);
    return 0 ;
  }
#endif

  /* check service name chars */
  while (*p != '\0') {
    if (!isdigit (*p) &&
	!islower (*p) &&
	!isupper (*p) &&
	strchr (PSVCNAME_XCHARS, *p) == 0) {
      errno = errnum (PSVC_ILLCHAR_IN_SVC);
      return 0 ;
    }
    p ++ ;
  }
  return 1;
}


static char*
normalize_fmt
  (const char *in,
   unsigned   len,
   char      *out)
{
  char *ret = out;

  /* syntax: [<number>] <type-letter> ... */
  while (len > 0) {
    int num = 1 ;

    /* skip blanks, comma colon and the like */
    while (isspace (in [0]) || ispunct (in [0])) {
      ++ in; 
      if (--  len == 0)
	goto done;
    }

    /* read a leading multiplier */
    if (isdigit (in [0])) {
      char *p;
      if ((num = strtol (in, &p, 10)) > MAX_FMT_MULTIPLIER) {
	errno = errnum (PSVC_FMTSIZE_2LARGE) ;
	return 0;
      }
      len -= (p - in) ;
      in = p ;
    }
    
    /* check type specifier */
    if (strchr ("LHCSD", in [0]) == 0) {
      errno = errnum (PSVC_FMT_NO_TYPE) ;
      return 0;
    }

    /* append type spec */
    if (num != 1) {
      sprintf (out, "%u%c", num, in [0]) ;
      out += strlen (out) ;
    } else
      * out ++ = * in ;

    ++  in ;
    -- len ;
  }

 done:
  * out = '\0' ;
  return ret;
}


/* ---------------------------------------------------------------------------- *
 *                      public rpc server functions                             *
 * ---------------------------------------------------------------------------- */


prpc_ftab*
psvc_set_proto
  (prpc_ftab   *rpc,
   int         (**Fn)(void*,void**),
   const char   *svc,
   const char   *fmt,
   int         laxen,
   void        **Env)
{
  void       *env = (Env != 0 ? *Env : 0);
  prpc_ftab **Rpc = &rpc ;
  
  char *p ;
  int n ;
  int (*fn)(void*,void**) ;
  svc_proto *pl, **Pl ;
  prpc_ftab *new ;


  /* check arguments */
  if (Fn == 0) {
    errno = errnum (PSVC_INT_ARG_ERROR);
    return 0;
  }
  if (!valid_psvc_name (svc))
    return 0 ;

  /* normalize arguments */
  if (fmt == 0) fmt = "" ;
  n = strlen (fmt) ;
  p = ALLOCA (n + 1) ;
  if ((fmt = normalize_fmt (fmt, n, p)) == 0) {
    DEALLOCA (p);
    return 0 ;
  }

  if (laxen != 0) laxen = 1 ;
    
  /* deallocate uopn error */
  new = (rpc == 0 ? (rpc = XMALLOC (sizeof (prpc_ftab))) : 0) ;
  fn  = *Fn ;
  
  if (new == 0) do {		/* find a matching svc */
    
    if (strcmp (rpc->svc, svc) != 0) 
      continue ;
    
    /* got entry, find a matching prototype in the prototype list */
    if (((void**)Pl = &rpc->plist, pl = *Pl) != 0) do {	
      
      /* overwrite or remove an existing descriptor */
      if (strcmp (pl->fmt, fmt) == 0 && pl->laxen == laxen) {
	  
	if (fn != 0)			/* overwrite ? */
	  goto overwrite_proto_and_return ;
	  
	XFREE (pl->fmt) ;		/* remove that proto type record */
	*Env = pl->env ;		/* pass back */
	*Fn  = pl->exec ;
	*Pl  = pl->next ;		/* unlink */
	
	if (pl == rpc->plist && pl->next == 0) {
	  XFREE (rpc->svc) ;		/* remove that svc record */
	  *Rpc = rpc->next ;		/* unlink */
	}
	goto return_one ;
	
      } /* end if overwrite or remove */

    } while (Pl = &pl->next, (pl = *Pl) != 0) ;
    
    /* append new prototype for that function: *Pl == <last-link> */
    goto set_proto_and_return;
    
  } while (Rpc = &rpc->next, (rpc = *Rpc) != 0) ;	
  
  /* end do: find a matching svc => no such function, yet */
  
  /* no list => destroy: nothing to do, here */
  if (fn == 0) goto return_null ;

  /* append new function and prototype, *Rpc == <last-link> */
  *Rpc = rpc  = XMALLOC (sizeof (prpc_ftab)) ;
  rpc->svc   = XSTRDUP (svc) ;  
  (void**)Pl = &rpc->plist ;

 set_proto_and_return:

  /* no list => destroy: nothing to do, here */
  if (fn == 0) goto return_null ;

  /* append new prototype: *Pl == <last-link> */
  *Pl = pl = XMALLOC (sizeof (svc_proto)) ;
  pl->fmt  = XSTRDUP (fmt) ;

 overwrite_proto_and_return:
  
  pl->laxen = laxen ;
  *Env      = pl->env ;
  pl->env   = env ;
  *Fn       = pl->exec ;
  pl->exec  = fn ;

 return_one:
  DEALLOCA (fmt) ;
  return rpc ;

 return_null:
  DEALLOCA (fmt) ;
  if (new != 0)
    XFREE (new) ;
  return 0 ;
}

/* ---------------------------------------------------------------------------- *
 *                      public rpc data store functions                         *
 * ---------------------------------------------------------------------------- */

prpc_data *
psvc_put_longx
  (prpc_data *rpc,
   long      *arg,
   unsigned   dim)
{
  prpc_data *new = (rpc == 0 ? (rpc = XMALLOC (sizeof (prpc_data))) : 0) ;
  char *trg ;

  if ((trg = append2rpc_spooler_prep (rpc, dim, "L", dim * sizeof (long))) == 0) {
    if (new != 0)
      XFREE (new) ;
    return 0 ;
  }
  
  /* transfer data */
  while (dim --) {
    ulong2buffer (trg, *arg) ;
    arg ++ ;
    trg += sizeof (long) ;
  }
  return rpc;
}


prpc_data *
psvc_put_shortx
  (prpc_data *rpc,
   short     *arg,
   unsigned   dim)
{
  prpc_data *new = (rpc == 0 ? (rpc = XMALLOC (sizeof (prpc_data))) : 0) ;
  char *trg ;

  if ((trg = append2rpc_spooler_prep (rpc, dim, "H", dim * sizeof (short))) == 0) {
    if (new != 0)
      XFREE (new) ;
    return 0 ;
  }
  
  /* transfer data */
  while (dim --) {
    ushort2buffer (trg, *arg) ;
    arg ++ ;
    trg += sizeof (short) ;
  }
  return rpc;
}


prpc_data *
psvc_put_charx
  (prpc_data *rpc,
   char      *arg,
   unsigned  size)
{
  prpc_data *new = (rpc == 0 ? (rpc = XMALLOC (sizeof (prpc_data))) : 0) ;
  char *trg ;

  if (size == 0)
    size = strlen (arg) + 1;
  
  if ((trg = append2rpc_spooler_prep (rpc, size, "C", size)) == 0) {
    if (new != 0)
      XFREE (new) ;
    return 0;
  }
  
  /* transfer data */
  if (size != 0)
    memcpy (trg, arg, size);
  
  return rpc;
}


prpc_data *
psvc_put_stringx
  (prpc_data *rpc,
   char     **arg,
   unsigned   dim)
{
  prpc_data *new = (rpc == 0 ? (rpc = XMALLOC (sizeof (prpc_data))) : 0) ;
  int size = 0 ;

  int n ;
  char *trg ;

  /* calculate legths */
  for (n = 0; n < dim; ++ n) {
    if (arg [n] == 0) {
      size += 2 ;
      continue ;
    }
    size += strlen (arg [n]) + 3;
  }
  
  if ((trg = append2rpc_spooler_prep (rpc, dim, "S", size)) == 0) {
    if (new != 0)
      XFREE (new) ;
    return 0 ;
  }

  /* transfer data */
  for (n = 0; n < dim; ++ n) {
    int m;

    /* store NULL string: <(short)NULL> */
    if (arg [n] == 0) {
      * trg ++ = 0 ;
      * trg ++ = 0 ;
      continue ;
    }

    /* store string data: <short-length> <string> '\0' */
    m = strlen (arg [n]) + 1;
    ushort2buffer (trg, m) ; trg += 2;
    memcpy (trg, arg [n], m) ; trg += m;
  }

  return rpc;
}



prpc_data *
psvc_put_datax
  (prpc_data *rpc,
   prpc_data *arg,
   unsigned   dim)
{
  prpc_data *new = (rpc == 0 ? (rpc = XMALLOC (sizeof (prpc_data))) : 0) ;
  int size = 0 ;

  int n, rpc_size, rpc_fmtlen ;
  char *trg ;

  /* some clown might choose rpc == arg */
  rpc_size   = rpc->size ;
  rpc_fmtlen = (rpc->fmt == 0 ? 0 : strlen (rpc->fmt)) ;

  /* calculate legths */
  for (n = 0; n < dim; ++ n) {
    int v = (arg [n].fmt == 0 ? 0 : strlen (arg [n].fmt)) ;
    if (v == 0 || arg [n].size == 0) {
      size += 2 ;
      continue ;
    }
    size += 3 + v + arg [n].size ;
  }

  if ((trg = append2rpc_spooler_prep (rpc, dim, "D", size)) == 0) {
    if (new != 0)
      XFREE (new) ;
    return 0 ;
  }

  /* transfer data */
  for (n = 0; n < dim; ++ n) {
    int m, v ;

    v = (rpc == arg 
	 ? rpc_fmtlen 
	 : (arg [n].fmt == 0 ? 0 : strlen (arg [n].fmt))) ;
    
    /* store NULL data: <(short)NULL> */
    if (v == 0 || arg [n].size == 0) {
      * trg ++ = 0 ;
      * trg ++ = 0 ;
      continue ;
    }

    /* store data: <short-length> <fmt> '\0' <data ...> */
    m = 3 + v + (rpc == arg ? rpc_size : arg [n].size) ;
    
    ushort2buffer (trg, m) ; trg += 2;
    memcpy (trg, arg [n].fmt, v) ; trg += v ;
    * trg ++ = '\0' ;
    if (arg [n].size != 0)
      memcpy (trg, arg [n].data, arg [n].size) ;
    trg += arg [n].size ;
  }

  return rpc;
}

/* ---------------------------------------------------------------------------- *
 *                  public rpc data clear/destroy functions                     *
 * ---------------------------------------------------------------------------- */

void
psvc_rewind
  (prpc_data *rpc)
{
  rpc->dataptr =
    rpc->fmtptr = 0 ;
}

void
psvc_clear
  (prpc_data *rpc)
{
  psvc_rewind  (rpc) ;

  if (rpc->data != 0) {
    XFREE (rpc->data) ;
    rpc->data = 0 ;
  }
  rpc->size = 0 ;
  
  if (rpc->fmt != 0) {
    XFREE (rpc->fmt) ;
    rpc->fmt = 0 ;
  }

  if (rpc->garbage != 0)
    yfree (rpc, 0);

  /* reset clear flag */
  rpc->clear_flag = 0 ;
}

void
psvc_destroy
  (prpc_data *rpc)
{
  psvc_clear (rpc);
  XFREE (rpc) ;
}

/* ---------------------------------------------------------------------------- *
 *                puplic argument data export/import                            *
 * ---------------------------------------------------------------------------- */

void *
psvc_export
  (prpc_data *rpc)
{
  char *p ;
  int n, len ;
  
  /* check arguments */
  if (rpc == 0) {
    errno = errnum (PRPC_INT_ARG_ERROR);
    return 0;
  }

  /* empty data set: <NULL> */
  if (rpc->data == 0 || rpc->fmt == 0 || rpc->fmt [0] == '\0')
    return XMALLOC (2) ;
  
  n  = strlen (rpc->fmt) + 1 ;
  if ((len = n + rpc->size) > MAX_FMT_MULTIPLIER - 2) {
    errno = errnum (PSVC_SIZE_2LARGE) ;
    return 0 ;
  }

  p = XMALLOC (len + 2) ;

  /* export format: <size> <format> <data ...> */
  ushort2buffer (p, len) ;
  memcpy (p + 2,     rpc->fmt, n);
  memcpy (p + 2 + n, rpc->data, rpc->size) ;

  return p;
}


unsigned
psvc_exportlen
  (const void *data)
{
  return data != 0 ? buffer2ushort ((char*)data) : 0 ;
}


prpc_data *
psvc_import
  (prpc_data   *rpc,
   const void *data)
{
  int           n = 0;
  const char *fmt = 0;

  int  len ;

  /* check arguments */
  if (data == 0) {
    errno = errnum (PRPC_INT_ARG_ERROR);
    return 0;
  }

  if ((len = buffer2ushort ((char*)data)) != 0) {
    fmt = (((char*)data) += 2) ;

    /* check strlen */
    for (n = 0; ((char*)data) [n] != 0; n ++)
      if (n >= len) {
	errno = errnum (PSVC_FMT_ERROR) ;
	return 0 ;
      }
    ++ n ;
  }

  if (rpc == 0) 
    rpc = XMALLOC (sizeof (prpc_data)) ;
  else
    psvc_clear (rpc) ;

  if (fmt != 0) {
    rpc->fmt  = XSTRDUP (fmt) ;
    if ((rpc->size = len - n) != 0) {
      rpc->data = XMALLOC (rpc->size) ;
      memcpy (rpc->data, fmt + n, rpc->size);
    }
  }

  return rpc ;
}


prpc_data *
psvc_cpy_data
  (prpc_data       *trg,
   const prpc_data *src)
{
  /* check arguments */
  if (src == 0) {
    errno = errnum (PRPC_INT_ARG_ERROR);
    return 0;
  }
  if (trg == 0) 
    trg = XMALLOC (sizeof (prpc_data)) ;
 
  *trg = *src ;
  
  if (trg->fmt != 0)
    trg->fmt = XSTRDUP (src->fmt) ;
  if (trg->data != 0)
    memcpy (trg->data = XMALLOC (trg->size), src->data, src->size);

  return trg;
}

/* ---------------------------------------------------------------------------- *
 *                  public rpc call & data retriever functions                  *
 * ---------------------------------------------------------------------------- */

int 
psvc_callx
  (const prpc_ftab *rpc,
   const char      *svc,
   prpc_data       *arg)
{
  svc_proto *pl_min =        0 ;
  int        ln_min = UINT_MAX ;
  
  int n, len ;
  svc_proto *pl ;
  void     *env ;

  /* check arguments */
  if (svc == 0 || rpc == 0 || arg == 0 || arg->fmt == 0) {
    errno = errnum (PRPC_INT_ARG_ERROR);
    return -1;
  }

  len = strlen (arg->fmt) ;

  if (rpc == 0) 
    goto no_such_function;
  
  /* find a matching svc */    
  while (strcmp (rpc->svc, svc) != 0)
    if ((rpc = rpc->next) == 0) 
      goto no_such_function;
  
  /* got svc, find a matching prototype */
  if ((pl = rpc->plist) == 0) 
    goto no_such_function;
    
  do {		/* loop through the prototype list */
      
    if (strncmp (pl->fmt, arg->fmt, n = strlen (pl->fmt)) != 0) 
      continue;
      
    /* does it match exactly ? */
    if (n == len) {
      
      if (pl->laxen == 0)
	/* pl->fmt was strict => match */
	break ;
      
      /* set the smallest matching pl->fmt */
      pl_min = pl ;
      ln_min =  n ;
      continue ;
    }
      
    /* pl->fmt is introduced by fmt */
    if (pl->laxen == 0)
      /* pl->fmt was strict => no match */
      continue ;
    
    /* find the smallest matching pl->fmt */
    if (ln_min > n) {
      pl_min = pl ;
      ln_min =  n ;
    }
	
  } while ((pl = pl->next) != 0) ;
      
  if (pl == 0 && (pl = pl_min) == 0)
    goto no_such_function;
  
  if (pl->exec == 0) 
    goto no_such_function;

  /* prpare for calling */
  errno = 0 ;

  /* fet flag to detect when clean up fn was called */
  arg->clear_flag = 1 ;

  /* call svc function */
  env     = pl->env ;
  n       = (*pl->exec)(arg, &env) ;
  pl->env = env ;

  /* clean up unless done by svc function */
  if (arg->clear_flag != 0)
    psvc_clear (arg);

  return n;
  
 no_such_function:
  errno = errnum (PSVC_NO_PROTOTYPE) ;
  return -1;
}



void *
psvc_call
  (const prpc_ftab *proto,
   const char         *svc,
   const void        *data)
{
  prpc_data *arg = psvc_import (0, data);

  int n ;

  if (arg == 0) 
    return 0 ;

  n    = psvc_callx (proto, svc, arg) ;
  data = (n < 0 ? 0 : psvc_export (arg)) ;

  psvc_destroy (arg);
  return (void*)data ;
}



int 
psvc_get_nextx
  (prpc_data  *rpc,
   void      **Arg,
   unsigned   *Dim)
{
  unsigned long dim = 1 ;

  long *larg;
  short *sarg;
  int n, m, size;
  char c, *p, *f, **argv, **av ;
  prpc_data *argd, *ad;
  garbage *stop ;

  /* nothing to do ? */
  if (rpc == 0 || rpc->data == 0 || rpc->fmt == 0 || rpc->fmt [0] == '\0')
    return 0 ;

  if (rpc->dataptr == 0) {
    rpc->dataptr = rpc->data ;
    rpc->left    = rpc->size ;
  }

  if (rpc->fmtptr == 0)
    rpc->fmtptr = rpc->fmt ;

  f = rpc->fmtptr ;
  
  /* get format spec */
  if (isdigit (f [0]) && (dim = strtol (f, &f, 10)) > MAX_FMT_MULTIPLIER) {
    errno = errnum (PSVC_SIZE_2LARGE) ;
    return 0 ;
  }
  /* adjust from strtol () */
  rpc->fmtptr = f ;

  /* needed for error recovery */
  stop = rpc->garbage ;

  switch (c = rpc->fmtptr [0]) {

  case 'L': /* long */

    if ((*Dim = dim) == 0) 
      break;

    if ((size = dim * sizeof (long)) > rpc->left) {
      errno = errnum (PSVC_FMT_ERROR) ;
      return -1 ;
    }

    *Arg = larg = yalloc (rpc, size) ;
    while (dim -- ) {
      *larg = buffer2ulong (rpc->dataptr) ;
      rpc->dataptr += sizeof (long) ;
      rpc->left    -= sizeof (long) ;
      larg ++ ;
    }
    break ;

  case 'H': /* short */

    if ((*Dim = dim) == 0) 
      break;

    if ((size = dim * sizeof (short)) > rpc->left) {
      errno = errnum (PSVC_FMT_ERROR) ;
      return -1 ;
    }

    *Arg = sarg = yalloc (rpc, size) ;
    while (dim -- ) {
      *sarg = buffer2ushort (rpc->dataptr) ;
      rpc->dataptr += sizeof (short) ;
      rpc->left    -= sizeof (short) ;
      sarg ++ ;
    }
    break ;

  case 'C': /* char */

    if ((*Dim = dim) == 0) 
      break;

    if (dim > rpc->left) {
      errno = errnum (PSVC_FMT_ERROR) ;
      return -1 ;
    }

    memcpy (*Arg = yalloc (rpc, dim), rpc->dataptr, dim) ;
    rpc->dataptr += dim ;
    rpc->left    -= dim ;
    break ;
    
  case 'S': /* text/string */
    
    if ((*Dim = dim) == 0) 
      break;

    argv = yalloc (rpc, sizeof (char*) * dim) ;
    p    = rpc->dataptr ;
    n    = rpc->left    ;
    
    av = argv ;
    while (dim -- ) {
      int len = buffer2ushort (p) ;

      /* undo everything on error */
      if (n < 2 + len) 
	goto data_pack_error;

      p += 2 ;
      n -= 2 ;
      if (len != 0) {
	if (p [len-1] != '\0') 
	  goto data_pack_error;
	/* let agrv point into the data aerea */
	*argv = p ;
	p += len ;
	n -= len ;
      }
      argv ++ ;
    }
    
    /* accept */
    rpc->dataptr =  p ;
    rpc->left    =  n ;
    *Arg         = av ;
    break ;

  case 'D': /* prpc data */

    if ((*Dim = dim) == 0) 
      break;

    argd = yalloc (rpc, dim * sizeof (prpc_data)) ;
    p    = rpc->dataptr ;
    n    = rpc->left    ;

    ad = argd ;
    while (dim -- ) {
      int len = buffer2ushort (p) ;

      /* undo everything on error */
      if (n < 2 + len)
	goto data_pack_error;
    
      p += 2 ;
      n -= 2 + len ;
      
      if (len != 0) {
	/* check strlen */
	for (m = 0; p [m] != 0; m ++)
	  if (m >= len) 
	    goto data_pack_error;
	m ++ ;
	memcpy (argd->fmt = yalloc (rpc, m), p, m) ; 
	p   += m ;
	len -= m ;
	memcpy (argd->data = yalloc (rpc, len), p, len);
	p   += len ;
      }
      argd ++ ;
    }

    /* accept */
    rpc->dataptr =  p ;
    rpc->left    =  n ;
    *Arg         = ad ;
    break ;

  default:
    errno = errnum (PSVC_FMT_ERROR) ;
    return -1 ;
  
  /*@NOTREACHED@*/

  data_pack_error:
    yfree (rpc, stop) ;
    errno = errnum (PSVC_FMT_ERROR) ;
    return -1 ;
  }

  if (dim == 0)
    *Arg = 0 ;

  rpc->fmtptr = f + 1;
  return c ;
}


char * 
psvc_get_proto
  (prpc_data *rpc)
{
  /* nothing to do ? */
  if (rpc->data == 0 || rpc->fmt == 0)
    return "" ;
  
  return (rpc->fmtptr == 0 ? rpc->fmt : rpc->fmtptr) ;
}

int
psvc_cmp_proto
  (prpc_data      *rpc,
   const char *pattern,
   unsigned      laxen)
{
  int n, test ;
  char *p, *q ;

  if (pattern == 0)
    return -1;
  
  p = psvc_get_proto (rpc) ;
  n = strlen (pattern) ;
  q = ALLOCA (n + 1) ;
  
  if ((pattern = normalize_fmt (pattern, n, q)) == 0) {
    DEALLOCA (q);
    return -1 ;
  }
  
  /* are there at least "pattern"-like prototype like args ? */
  test = strncmp (p, pattern, n = strlen (pattern)) ;
  DEALLOCA (q) ;

  return 
    /* return 0 on match, 1 otherwise */
    test != 0 || (n != strlen (p) && laxen == 0) ;
}


#if 1
/* ---------------------------------------------------------------------------- *
 *                      TEST                                                    *
 * ---------------------------------------------------------------------------- */

static char *envp = "test_value" ;

static int
test_log_1
  (void *data,
   void **env)
{
  return 0;
}

static int
test_log_2
  (void *data,
   void **env)
{
  return 0;
}

static int
test_log_3
  (void *data,
   void **env)
{
  /* prototype is "2SH" */
  int n1 =  psvc_cmp_proto (data, "2S,  ", 0);
  int n2 =  psvc_cmp_proto (data, "2S,  ", 1);
  int n3 =  psvc_cmp_proto (data, "2S, H", 0);
  int n4 =  psvc_cmp_proto (data, "2S, H", 1);
  int n5 =  psvc_cmp_proto (data, "2S, H, C", 1);

  short h;
  void *av;
  int dim, c, val ;
  char *channel, *format;

  c = psvc_get_nextx (data, &av, &dim) ;
  channel = ((char**)av) [0];
  format  = ((char**)av) [1];
  
  c = psvc_get_nextx (data, &av, &dim) ;
  val = ((short*)av) [0] ;


  psvc_clear (data);
  h = 33 ;
  psvc_put_shortx (data, &h, 1) ;
  return 0;
}


int
prpc_test () {
  void *ev;
  int (*fn)(void*,void**) ;
  short h ;
  char *p, *q, *s2 [2] ;
  prpc_ftab *rp = 0;
  prpc_data *r = 0, *u ;
    
  ev = envp ;
  fn = test_log_1 ;
  rp = psvc_set_proto (rp, &fn, "log", " S, S", 1, &ev) ;

  ev = envp ;
  fn = test_log_2 ;
  rp = psvc_set_proto (rp, &fn, "log", "2S, H", 1, &ev) ;

  ev = envp ;
  fn = test_log_3 ;
  rp = psvc_set_proto (rp, &fn, "log", "2S, H ", 1, &ev) ;

  s2 [0] = "stderr" ;
  s2 [1] = "haha" ;
  r      = psvc_put_stringx (r, s2, 2) ;
  h      = 77 ;
  r      = psvc_put_shortx  (r, &h, 1) ;

  p      = psvc_export (r) ;
  q      = psvc_call (rp, "log", p);
  u      = psvc_import (r, q) ;

#if 0
  n      = psvc_callx (rp, "log") ;
#endif

  return 0 ;
}
#endif
