/*
 * Copyright (C) 1999 Renaud Deraison
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * TCP Ping. An addition to the libpcap
 */

#include <includes.h>
#include <nessusraw.h>
#include <pcap.h>
struct pseudohdr
{
        struct in_addr saddr;
        struct in_addr daddr;
        u_char zero;
        u_char protocol;
        u_short length;
        struct tcphdr tcpheader;
};

int np_in_cksum(u_short *, int);
pcap_t * np_init_pcap(struct in_addr, struct in_addr, int);
unsigned char * np_forge_tcp_packet(struct in_addr, struct in_addr,
					unsigned int, unsigned int, 
					int, long);

/*
 * 
 * Initialize the libpcap
 *
 */
pcap_t *
np_init_pcap(src, dst, to)
 struct in_addr src;
 struct in_addr dst;
 int to;
{
 char * errbuf = emalloc(PCAP_ERRBUF_SIZE);
 pcap_t * ret;
 char * interface;
 char * a_dst;
 bpf_u_int32 netmask, network;
 struct bpf_program filter_prog;
 char * filter = emalloc(1024);
 a_dst = estrdup(inet_ntoa(dst));
 sprintf(filter, "src host %s",
 		  a_dst);
 efree(&a_dst);
 interface = routethrough(&dst, &src);
 if(!interface)interface = pcap_lookupdev(errbuf);
 if(pcap_lookupnet(interface, &network, &netmask, 0)<0)return(NULL);
 ret = pcap_open_live(interface, 92, 0, to, errbuf);
 if(pcap_compile(ret, &filter_prog, filter, 1, netmask)<0)return(NULL);
 if(pcap_setfilter(ret, &filter_prog)<0)return(NULL);		  
 
 return(ret);
}


/*
 *
 * Forge a TCP packet
 *
 */
unsigned char * 
np_forge_tcp_packet(source, dest, sport, dport, flags, seq)
 struct in_addr source;
 struct in_addr dest;
 unsigned int sport;
 unsigned int dport;
 int flags;
 long seq;
{
 u_char * ret = emalloc(sizeof(struct ip)+sizeof(struct tcphdr));
 struct pseudohdr pseudoheader;
 struct ip * ip = (struct ip *)ret;
 struct tcphdr * tcp = (struct tcphdr*)(ret + sizeof(struct ip));
 /*
  * Initialize the IP header
  */
 ip->ip_v = 4;
 ip->ip_hl = sizeof(struct ip)/4;
 ip->ip_id = random();
 ip->ip_tos = 0;
 ip->ip_len = FIX(sizeof(struct ip) + sizeof(struct tcphdr));
 ip->ip_ttl = 255;
 ip->ip_off = 0;
 ip->ip_p = IPPROTO_TCP;
 ip->ip_src.s_addr = source.s_addr;
 ip->ip_dst.s_addr = dest.s_addr;
 ip->ip_sum = 0;
 ip->ip_sum = np_in_cksum((short *)ip, sizeof(struct ip));
 
 tcp->th_sport = htons(sport);
 tcp->th_dport = htons(dport);
 tcp->th_seq = htonl(seq);
 tcp->th_ack = 0;
 tcp->th_flags = flags;
#ifdef HAVE_TCPHDR_TH_OFF
 tcp->th_x2 = 0;
 tcp->th_off = 5;
#else
 tcp->th_x2_off = 5 << 4;
#endif
 tcp->th_win = htons(4096);
 tcp->th_sum = 0;
 tcp->th_urp  = 0;
 
 bzero(&pseudoheader, 12+sizeof(struct tcphdr));
 pseudoheader.saddr.s_addr=source.s_addr;
 pseudoheader.daddr.s_addr=dest.s_addr;
 
 pseudoheader.protocol=6;
 pseudoheader.length=htons(sizeof(struct tcphdr));
 bcopy((char *) tcp,(char *) &pseudoheader.tcpheader,sizeof(struct tcphdr));
 tcp->th_sum = np_in_cksum((unsigned short *)&pseudoheader,12+sizeof(struct tcphdr));
 return(ret);
}

/*
 * 
 * Send a TCP ping
 *
 */
int 
np_tcp_ping(pcap, soc, src, dst, flag, port)
 pcap_t * pcap;
 int soc;
 struct in_addr src;
 struct in_addr dst;
 int flag;
 unsigned short port;
{
 char * packet = np_forge_tcp_packet(src, dst, 10101,port , flag, random());
 struct sockaddr_in socaddr;
 int b;
 int ret = 0;
 bzero(&socaddr, sizeof(socaddr));
 socaddr.sin_family = AF_INET;
 socaddr.sin_addr.s_addr = dst.s_addr;
 socaddr.sin_port = htons(11111);
 b = sendto(soc, packet, sizeof(struct ip)+sizeof(struct tcphdr), 0, 
 	(struct sockaddr *)&socaddr, sizeof(struct sockaddr_in));
 efree(&packet);
 if(b>0)
 {
  struct pcap_pkthdr head;
  int dl_len = get_datalink_size(pcap_datalink(pcap));
  packet = (char*)pcap_next(pcap, &head);
  if(packet)
  {
   struct ip * ip = (struct ip *)(packet + dl_len);
   if(ip->ip_src.s_addr == dst.s_addr)ret++;
  }
 }
 return(ret);
}


  
/*
 * tcp_ping_host()
 *
 * This function pings the remote host via TCP : it sends it a 
 * TH_SYN|TH_ACK packet (which bypasses most packet filters) 
 */
int
tcp_ping_host(addr)
 struct in_addr addr;
{
 pcap_t * pcap;
 struct in_addr src;
 char * name; 
 int soc;
 int ret = 0;
 int one = 1;
 
 soc = socket(AF_INET, SOCK_RAW, IPPROTO_RAW);
#ifdef IP_HDRINCL
 setsockopt(soc, IPPROTO_IP, IP_HDRINCL, &one, sizeof(one));
#endif
 if(soc < 0)return(0);
 name = emalloc(255);
 gethostname(name, 255);
 src = nn_resolve(name);
 efree(&name);
 if(src.s_addr == addr.s_addr)ret++;
 else
 {
  pcap = np_init_pcap(src,addr, 800);
  if(pcap)
  {
  if(!(ret =  np_tcp_ping(pcap, soc, src, addr,TH_SYN|TH_ACK, 80)))
     ret = np_tcp_ping(pcap, soc, src,addr,TH_FIN, 11111);
  }
  pcap_close(pcap);
 }
 shutdown(soc, 2);
 close(soc);
 return(ret);
}





/*
 * Checksum routine for Internet Protocol family headers (C Version)
 * From ping examples in W.Richard Stevens "UNIX NETWORK PROGRAMMING" book.
 */
int np_in_cksum(p,n)
u_short *p; int n;
{
  register u_short answer;
  register long sum = 0;
  u_short odd_byte = 0;

  while( n > 1 )  { sum += *p++; n -= 2; }

  /* mop up an odd byte, if necessary */
  if( n == 1 ) {
      *(u_char *)(&odd_byte) = *(u_char *)p;
      sum += odd_byte;
  }

  sum = (sum >> 16) + (sum & 0xffff);	/* add hi 16 to low 16 */
  sum += (sum >> 16);			/* add carry */
  answer = (int)~sum;			/* ones-complement, truncate*/
  return (answer);
}
