/*
 * Copyright (C) 1999 Renaud Deraison
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * TCP Timer. Another addition to the libpcap
 */
 
#include <includes.h>
#include <nessusraw.h>
#include <pcap.h>
#include "tcp_ping.h"




long tcp_timing_send(pcap, soc, src, dst, port)
 pcap_t * pcap;
 int soc;
 struct in_addr src, dst;
 unsigned int port;
{
 struct timeval then, now;
 struct timezone z;
 long ret;
 int b;
 char * packet;
 struct sockaddr_in sockaddr;
 int dl_len = get_datalink_size(pcap_datalink(pcap));
 
 bzero(&sockaddr, sizeof(struct sockaddr));
 sockaddr.sin_family = AF_INET;
 sockaddr.sin_addr.s_addr = dst.s_addr;
 sockaddr.sin_port = htons(10000);
 gettimeofday(&then, &z);
 ret = then.tv_sec * 1000;
 ret += then.tv_usec / 1000;
 packet = np_forge_tcp_packet(src, dst, 10000, port, TH_SYN, ret);
 b = sendto(soc, packet, sizeof(struct ip)+sizeof(struct tcphdr), 0, 
 	(struct sockaddr *)&sockaddr, sizeof(struct sockaddr_in));
 efree(&packet);
 ret = -1;
 if(b > 0)
 {
  struct pcap_pkthdr head;
  int j;
  for(j=0;j<80 && !packet;j++)packet = (char *)pcap_next(pcap, &head);
  if(packet)
  {
   long nw;
   struct ip * ip = (struct ip *)(packet + dl_len);
   struct tcphdr * tcp;
  
   tcp  = (struct tcphdr *)(packet + 20 + dl_len);
   if(ip->ip_p == IPPROTO_TCP)
   {
   gettimeofday(&now, &z);
   nw = now.tv_sec * 1000;
   nw += now.tv_usec / 1000;
   ret = nw - ntohl(tcp->th_ack);
   }
  }
 }
 return(ret);
}
 
long tcp_timing(addr, num, port)
 struct in_addr addr;
 int num;
 unsigned int port;
{
 pcap_t * pcap = NULL;
 int soc;
 char * name;
 int one = 1;
 struct in_addr src;
 long ret = -1;
 soc = socket(AF_INET, SOCK_RAW, IPPROTO_RAW);
#ifdef IP_HDRINCL
 setsockopt(soc, IPPROTO_IP, IP_HDRINCL, &one, sizeof(one));
#endif
 if(soc < 0)return(-1);
 name = emalloc(255);
 gethostname(name, 255);
 src = nn_resolve(name); 
 efree(&name);
 if(src.s_addr == addr.s_addr)ret = 0; /* localhost */
 else
 {
 pcap = np_init_pcap(src, addr,5);
 if(pcap)
  {
   int i;
   long average = 0;
   int num_elems = 0;
   long * results = emalloc(sizeof(long)*(num+1));
   for(i=0;i<num;i++)results[i] = tcp_timing_send(pcap, soc, src, addr, port);
   for(i=0;i<num;i++)if(results[i]>=0){
   	average+=results[i];
	num_elems++;
	}
   ret = average/num_elems;
   pcap_close(pcap);
  }
 }
 close(soc);
 return(ret);
}
