/* Nessus
 * Copyright (C) 1998 Renaud Deraison
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
 
#include <includes.h>
#ifndef NESSUSNT
#ifdef HAVE_GETOPT_H
#include <getopt.h>
#else  
extern int optind;
#endif /* not HAVE_GETOPT_H */   

#else/* not def NESSUSNT */
#include "wstuff.h"
#endif

#ifdef ENABLE_CRYPTO_LAYER
#include "iostream.h"
#include "peks.h"
#include "nsp.h"
#endif


#include "pluginload.h"
#include "preferences.h"
#include "auth.h"
#include "rules.h"
#include "comm.h"
#include "attack.h"
#include "sighand.h"
#include "log.h"
#include "threads.h"
#include "users.h"
#include "ntp.h"
#include "utils.h"

#ifdef DEBUG
int single_shot = 0 ;
#endif

/*
 * Functions prototypes
 */
static void main_loop(int, struct arglist *, struct arglist *, struct nessus_rules*, 
	     struct users *);
static int init_nessusd(struct arglist *);
static int init_network(int, int *, struct in_addr);
static void server_thread(struct arglist *);


static void 
server_thread(globals)
 struct arglist * globals;
{
 struct sockaddr_in * address = arg_get_value(globals, "client_address");
 struct arglist * plugins = arg_get_value(globals, "plugins");
#ifdef ENABLE_CRYPTO_LAYER
 struct arglist * prefs = arg_get_value (globals, "preferences") ;
 char * opt = NULL;
#endif
 int soc = (int)arg_get_value(globals, "global_socket");
 char * perms;
 char * asciiaddr;
 struct nessus_rules * rules = arg_get_value(globals, "rules");
 ntp_caps* caps;
 
#ifdef USE_PTHREADS
 int off = 0;
 ioctl(soc, FIONBIO, &off);
#endif

#ifdef HAVE_ADDR2ASCII
 asciiaddr = emalloc(20);
 addr2ascii(AF_INET, &address->sin_addr, sizeof(struct in_addr), asciiaddr);
#elif defined(HAVE_INET_NETA)
 asciiaddr = emalloc(20);
 inet_neta(ntohl(address->sin_addr.s_addr), asciiaddr, 20);
#else
 asciiaddr = estrdup(inet_ntoa(address->sin_addr));
#endif
 caps = comm_init(soc);
#ifdef ENABLE_CRYPTO_LAYER
 opt = arg_get_value(prefs, "force_pubkey_auth");
 if(opt && !strncmp(opt, "yes", 3) && !caps->pubkey_auth)
  {
   char * error = emalloc(256);
   log_write("%s wants to connect without pubkey auth\n", 
   		(char*)asciiaddr);		
   sprintf(error, "This server only accepts public-key \
authentificated users");
    send(soc, error, strlen(error), 0);
    efree(&error);
    goto shutdown_and_exit;
   }
 if (caps->ciphered) {
   int n;
   if (server_negotiate_session_key 
	(arg_get_value (prefs, "peks_key"),		/* default peks key */
	 soc,						/* tcp session */
	 asciiaddr,    			                /* client address */
	 arg_get_value (prefs, "peks_keyfile")) < 0) {	/* cipher and key list file */
     log_write ("While negotiating key with %s: %s\n",
		asciiaddr, peks_strerr (errno)) ;		
     goto shutdown_and_exit;
   }
  n = 4096;
  io_ctrl (soc, IO_RESIZE_BUF, &n, 1) ; /* write buffer */
 }
#endif /* ENABLE_CRYPTO_LAYER */

 arg_add_value(globals, "ntp_caps", ARG_STRUCT, sizeof(*caps), caps);
 plugins_set_ntp_caps(plugins, caps);
 plugins_set_socket(plugins, soc);
 if(!(perms = auth_check_user(globals, asciiaddr)))
 {
   auth_printf(globals, "Bad login attempt !\n"); 
   log_write("bad login attempt from %s\n", 
			asciiaddr);
   efree(&asciiaddr);			
 }
  else {
   efree(&asciiaddr);
   rules_add(rules, perms, 1);
   comm_send_pluginlist(globals);
   if(caps->ntp_11){
       comm_send_preferences(globals);
       comm_send_rules(globals);
       }
   comm_wait_order(globals);
   attack_host(globals);
   comm_terminate(globals);
   }
#ifdef ENABLE_CRYPTO_LAYER
 shutdown_and_exit:
#endif
 shutdown(soc, 2);
 close(soc);
#ifndef NESSUSNT
 arg_free(globals);
#endif
 EXIT(0);
}
     
static void 
main_loop(serv_socket, plugins, preferences, rules, users)
 int serv_socket;
 struct arglist * plugins;
 struct arglist * preferences;
 struct nessus_rules * rules;
 struct users * users;
{
  while(1)
    {
      int soc;
      unsigned int lg_address = sizeof(struct sockaddr_in);
      struct sockaddr_in address;
  
      struct arglist * globals;
      struct arglist * my_plugins, * my_preferences, * my_scanners;
      struct nessus_rules * my_rules;
      struct sockaddr_in * p_addr;
      struct arglist * plugs;
      struct arglist * scanners;
      char * asciiaddr;
      soc = accept(serv_socket, (struct sockaddr *)(&address), &lg_address);
      if(soc == -1){
#ifndef NESSUSNT
    	if(errno == EINTR)continue;
  	else 
#endif
	{
	  perror("accept ");
	  DO_EXIT(2);
	}
      }
#ifdef HAVE_ADDR2ASCII
      asciiaddr = emalloc(20);
      addr2ascii(AF_INET, &address.sin_addr, sizeof(struct in_addr), asciiaddr);
#elif defined(HAVE_INET_NETA)
      asciiaddr = emalloc(20);
      inet_neta(ntohl(address.sin_addr.s_addr), asciiaddr, 20);
#else      
      asciiaddr = estrdup(inet_ntoa(address.sin_addr));
#endif            
      log_write("connection from %s\n", (char *)asciiaddr);
      efree(&asciiaddr);
      /* 
       * duplicate everything so that the threads don't share the
       * same variables.
       *
       * Useless when fork is used, necessary for the pthreads and
       * the NT threads
       */
      globals = emalloc(sizeof(struct arglist));
      arg_add_value(globals, "global_socket", ARG_INT, -1, (void *)soc);
          

      my_plugins = emalloc(sizeof(struct arglist));
      plugs = arg_get_value(plugins, "plugins");
      arg_dup(my_plugins, plugs);
      arg_add_value(globals, "plugins", ARG_ARGLIST, -1, my_plugins);
      
      my_scanners = emalloc(sizeof(struct arglist));
      scanners = arg_get_value(plugins, "scanners");
      arg_dup(my_scanners, scanners);
      arg_add_value(globals, "scanners", ARG_ARGLIST, -1, my_scanners);
          
      my_preferences = emalloc(sizeof(struct arglist));
      arg_dup(my_preferences, preferences);
      arg_add_value(globals, "preferences", ARG_ARGLIST, -1, my_preferences);
          
      my_rules = emalloc(sizeof(struct nessus_rules));
      rules_dup(my_rules, rules);
      
      arg_add_value(globals, "users", ARG_PTR, -1, users);
          
      p_addr = emalloc(sizeof(struct sockaddr_in));
      *p_addr = address;
      arg_add_value(globals, "client_address", ARG_PTR, -1, p_addr);
      
      arg_add_value(globals, "rules", ARG_PTR, -1, my_rules);
      /* we do not want to create an io thread, yet so the last argument is -1 */
      create_thread(server_thread, globals, (struct arglist *)(-1));
    }
}


/*
 * Initialization of the network : 
 * we setup the socket that will listen for
 * incoming connections on port <port> on address
 * <addr> (which are set to 3001 and INADDR_ANY by
 * default)
 */ 
static int 
init_network(port, sock, addr)
     int port;
     int * sock;
     struct in_addr addr;
{

#ifndef NESSUSNT
  int option = 1;
#endif

#ifdef USE_AF_INET
  struct sockaddr_in address;
#else
  struct sockaddr_un address;
  char * name = AF_UNIX_PATH;
#endif


#ifdef NESSUSNT
    WORD wVersionRequested;
    WSADATA wsaData;
    int err;
    
    wVersionRequested = MAKEWORD( 2, 0 );
    err = WSAStartup( wVersionRequested, &wsaData );
    if(err){print_error("Could not initialize WinSock !");DO_EXIT(0);}
#endif

#ifdef USE_AF_INET
  if((*sock = socket(AF_INET, SOCK_STREAM, 0))==-1)
    {
      perror("socket ");
      DO_EXIT(1);
    }
  bzero(&address, sizeof(struct sockaddr_in));
  address.sin_family = AF_INET;
  address.sin_addr = addr;
  address.sin_port = htons((unsigned short)port);
#else
  if((*sock = socket(AF_UNIX, SOCK_STREAM,0))==-1)
    {
     perror("socket ");
     DO_EXIT(1);
   }
  bzero(&address, sizeof(struct sockaddr_un));
  address.sun_family = AF_UNIX;
  bcopy(name, address.sun_path, strlen(name));
#endif

#ifndef NESSUSNT
  setsockopt(*sock, SOL_SOCKET, SO_REUSEADDR, &option, sizeof(int));
#endif
  if(bind(*sock, (struct sockaddr *)(&address), sizeof(address))==-1)
    {
      perror("bind ");
      DO_EXIT(1);
    }
  if(listen(*sock, 10)==-1)
    {
      perror("listen ");
      shutdown(*sock, 2);
      close(*sock);
      DO_EXIT(1);
    }
  return(0);
}

/*
 * Initialize everything
 */
static int 
init_nessusd(options)
     struct arglist * options;    
    
  
{
  int  sock;
  struct arglist * plugins;
  struct arglist * preferences;
  struct nessus_rules * rules;
  struct users * users;
  int serv_port = (int)arg_get_value(options, "serv_port");
  char * config_file = arg_get_value(options, "config_file");
  struct in_addr * addr = arg_get_value(options, "addr");
#ifdef ENABLE_CRYPTO_LAYER
  char *pwd_user = arg_get_value(options, "pwd_user");
  char *pwd_passwd = arg_get_value(options, "pwd_passwd");
  char *kill_userkey = arg_get_value(options, "kill_userkey");
  int list_udb = (int)arg_get_value(options, "list_udb");
#endif
  preferences_init(config_file, &preferences);

#ifdef ENABLE_CRYPTO_LAYER
  {
    peks_key *key;

    /* overwrite defaults by config */
    if (arg_get_value (preferences, "peks_keylen") == 0) {
      char buf [20]; sprintf (buf, "%u", NESSUSD_KEYLENGTH) ;
      arg_add_value (preferences, "peks_keylen", ARG_STRING, strlen (buf), buf);
    }
 
    if (arg_get_value (preferences, "peks_keyfile") == 0)
      arg_add_value (preferences, "peks_keyfile", ARG_STRING, 
		     strlen (NESSUSD_KEYFILE), NESSUSD_KEYFILE);
    
    if (arg_get_value (preferences, "peks_username") == 0)
      arg_add_value (preferences, "peks_username", ARG_STRING, 
 		     strlen (NESSUSD_USERNAME), NESSUSD_USERNAME);
     
    if (arg_get_value (preferences, "peks_usrkeys") == 0)
      arg_add_value (preferences, "peks_usrkeys", ARG_STRING, 
 		     strlen (NESSUSD_USRKEYS), NESSUSD_USRKEYS);
       
    if ((key = peks_private_key 
 	 (arg_get_value (preferences, "peks_username"),
 	  arg_get_value (preferences, "peks_keyfile"),
 	  atoi (arg_get_value (preferences, "peks_keylen")))) == 0) {
       print_error("%s - aborting.\n", peks_strerr (errno));
       DO_EXIT (1);
    }

    if (pwd_user != 0 || kill_userkey != 0 || list_udb) 
      /* check whether we are allowed to do that */
      if (getuid () != 0) {
 	/* we check the real id despite the fact that nessusd usually runs 
 	   suid root */
 	print_error("Only root may access the user data base - aborting.\n");
 	DO_EXIT (1);
      }
    
    /* password table edit request */
    if (pwd_user != 0) {
      char * s;
 
      /* delete entry ? */
      if (pwd_passwd != 0 && pwd_passwd [0] == 0) {
 	pwd_passwd = 0; 
 	key = 0;
      }
 
      if ((s = peks_edit_passwd_file
 	   (key, pwd_user, 0, pwd_passwd,
 	    arg_get_value (preferences, "peks_usrkeys"))) == 0) {
 	print_error("ERROR: %s.\n", peks_strerr (errno));
 	DO_EXIT (1);
      }
 
      if (s [0])
 	printf ("%s\n", s) ;
    }
 
    /* user key table delete request */
    if (kill_userkey != 0) {
      if (peks_delete_userkey 
 	  (kill_userkey, arg_get_value (preferences, "peks_usrkeys")) < 0) {
 	print_error("ERROR: %s.\n", peks_strerr (errno));
 	DO_EXIT (1);
      }
    }
 
    /* user key table list request */
    if (list_udb != 0) {
      if (peks_list_keyfile (0 /* default fn prints to stdout */, 
 			     arg_get_value (preferences, "peks_usrkeys")) < 0) {
 	print_error("ERROR: %s.\n", peks_strerr (errno));
 	DO_EXIT (1);
      }
    }
 
    /* no connection, but user passwd table etc. request */
    if (pwd_user != 0 || kill_userkey != 0 || list_udb)
      DO_EXIT (0) ;

    /* store key for global use */
    arg_add_value(preferences, "peks_key", ARG_PTR, -1, key);
  }
#endif /* ENABLE_CRYPTO_LAYER */

  arg_add_value(preferences, "config_file", ARG_STRING, strlen(config_file), 
                config_file);
  log_init(arg_get_value(preferences, "logfile"));
  rules_init(&rules, preferences);
  users_init(&users, preferences);
  plugins = plugins_init(preferences);
  init_network(serv_port, &sock, *addr);
#ifndef USE_PTHREADS
  signal(SIGCHLD, sighand_chld);
  signal(SIGTERM, sighand_term);
  signal(SIGINT, sighand_int);
  signal(SIGKILL, sighand_kill);
#ifndef DEBUG
  signal(SIGSEGV, sighand_segv);
#endif
#endif
   arg_add_value(options, "sock", ARG_INT, sizeof(int),(void *)sock);
   arg_add_value(options, "plugins", ARG_ARGLIST, -1, plugins);
   arg_add_value(options, "rules", ARG_PTR, -1, rules);
   arg_add_value(options, "users", ARG_PTR, -1, users);
   arg_add_value(options, "preferences", ARG_ARGLIST, -1, preferences);
  return(0);
}


static void 
display_help()
{
  printf("nessusd, version %s\n", NESSUS_VERSION);
#ifdef USE_AF_INET
#ifdef ENABLE_CRYPTO_LAYER
  printf("\nusage : nessusd [options]\n\noptions:\n");
#else
  printf("\nusage : nessusd [-vcphdD] [-a address]\n\n");
#endif /* ENABLE_CRYPTO_LAYER */
#else
  printf("\nusage : nessusd [-vchdD]\n\n");
#endif /*def USE_AF_INET */

  printf("\ta <address>    : listen on <address>\n");
  printf("\tv              : shows version number\n");
  printf("\th              : shows this help\n");

#ifdef USE_AF_INET
  printf("\tp <number>     : use port number <number>\n");
#endif /* USE_AF_INET */
#ifdef ENABLE_CRYPTO_LAYER
  printf("\tP <user>       : list temporary user passwd\n");
  printf("\tP <user>,<pwd> : replace temporary user passwd\n");
  printf("\tP <user>,      : delete temporary user passwd\n");
  printf("\tK <user>       : kill registered user auth key\n");
  printf("\tL              : list user auth key entries\n");
 #endif /* ENABLE_CRYPTO_LAYER */
  printf("\tc <filename>   : alternate configuration file to use\n");
  printf("\t\t\t (default : %s)\n", NESSUSD_CONF);
  printf("\tD              : runs in daemon mode\n");
  printf("\td              : dumps the nessusd compilation options\n");
}

#ifdef NESSUSNT
int WINAPI WinMain(HINSTANCE hThisInst, HINSTANCE hPrevInst,
                                 LPSTR lpszArgs, int nWinMode)
#else
int 
main(int argc, char * argv[])
#endif
{
  int serv_port = -1;
  int sock;
  char * config_file = NULL;
  struct users * users;
  struct nessus_rules * rules;
  int do_fork = 0;
  struct arglist * plugins;
  struct arglist * preferences;
  struct in_addr addr; 
  struct arglist * options = emalloc(sizeof(struct arglist));
#ifdef ENABLE_CRYPTO_LAYER
  static char *pwd_user = NULL;
  static char *pwd_passwd = NULL;
  static char *kill_userkey = NULL;
  int list_udb = 0;
#endif  
#ifndef NESSUSNT
  int i;
#endif
  addr.s_addr = htonl(INADDR_ANY);
#ifndef NESSUSNT
  while((i=getopt(argc, argv, "Dhvc:d0123a:p:K:P:Le"))!=-1)
      switch(i)
	{
	case 'e' :
	  break;
	  
        case 'a' :
          if(!optarg){display_help();DO_EXIT(0);}
          if(!inet_aton(optarg, &addr)){display_help();DO_EXIT(0);}
          break;
          
	case 'p' :
	  if(!optarg){display_help();DO_EXIT(0);}
	  serv_port = atoi(optarg);
	  if((serv_port<=0)||(serv_port>=65536))
	    {
	      printf("Invalid port specification\n");
	      display_help();
	      DO_EXIT(1);
	    }
	  break;
	  
 	case 'P' :
#ifdef ENABLE_CRYPTO_LAYER
 	  /* syntax: user[,passwd] */
 	  pwd_user = estrdup (optarg) ;
 	  if ((pwd_passwd = strchr (pwd_user, ',')) != 0) 
 	    * pwd_passwd ++ = '\0' ;
 	  break;
#else
 	  goto disp_help ;
#endif /* ENABLE_CRYPTO_LAYER */
 
 	case 'K' :
#ifdef ENABLE_CRYPTO_LAYER
 	  /* syntax: user */
 	  kill_userkey = estrdup (optarg) ;
 	  break;
#else
 	  goto disp_help ;
#endif /* ENABLE_CRYPTO_LAYER */
 
 	case 'L' :
#ifdef ENABLE_CRYPTO_LAYER
	  list_udb++;
	  break;
#else
 	  goto disp_help ;
#endif /* ENABLE_CRYPTO_LAYER */
 
	case 'D' : 
	  do_fork++;
	  break;
	  
	case 'v' :
	  print_error("nessusd (%s) %s for %s\n(C) 1998-1999 Renaud Deraison <deraison@nessus.org>\n\n", 
		 PROGNAME,NESSUS_VERSION, NESS_OS_NAME);
	  DO_EXIT(0);
	  break;
	case '1' :
#ifdef DUMP_IOLAYER
	  dump_send = 1;
	  break;
#else
 	  goto disp_help ;
#endif
	case '0' :
#ifdef DUMP_IOLAYER
	  dump_recv = 1;
	  break;
#else
 	  goto disp_help ;
#endif
 	case '3' :
#ifdef DEBUG
 	  single_shot = 1;
	  break;
#endif

	case 'h' :
 	case '?' : /* getopt: error */
 	case ':' : /* getopt: missing parameter */
 	disp_help:
	  display_help();
	  DO_EXIT(0);
	  break;
	case 'c' : 
	  if(!optarg){display_help();DO_EXIT(1);}
	  config_file = emalloc(strlen(optarg)+1);
	  strncpy(config_file, optarg, strlen(optarg));
	  break;
       case 'd' :
           printf("This is Nessus %s for %s %s\n", NESSUS_VERSION, NESS_OS_NAME, NESS_OS_VERSION);
           printf("compiled with %s\n", NESS_COMPILER);
           printf("Current setup :\n");
           printf("\tThread manager : ");
#ifdef USE_PTHREADS
	   printf("pthreads\n");
#else
	   printf("fork\n");
#endif
#ifdef ENABLE_CRYPTO_LAYER
	   printf("\tCrypto layer : %s\n", peks_version ());
#endif
	   printf("\tRunning as euid : %d\n", geteuid());
           printf("\n\nInclude these infos in your bug reports\n");
	   DO_EXIT(0);
	   break;
	}
#endif
  if(serv_port == -1)serv_port = DEFAULT_PORT;
  if(!config_file)
    {
      config_file = emalloc(strlen(NESSUSD_CONF)+1);
      strncpy(config_file, NESSUSD_CONF, strlen(NESSUSD_CONF));
    }
  arg_add_value(options, "serv_port", ARG_INT, sizeof(int), (void *)serv_port);
  arg_add_value(options, "config_file", ARG_STRING, strlen(config_file), 
  		config_file);
  arg_add_value(options, "addr", ARG_PTR, -1, &addr);
#ifdef ENABLE_CRYPTO_LAYER
  arg_add_value(options, "pwd_user", ARG_PTR, -1, pwd_user);
  arg_add_value(options, "pwd_passwd", ARG_PTR, -1, pwd_passwd);
  arg_add_value(options, "kill_userkey", ARG_PTR, -1, kill_userkey);
  arg_add_value(options, "list_udb", ARG_INT, sizeof(int), (void*)list_udb);
#endif
  init_nessusd(options);
  sock = (int)arg_get_value(options, "sock");
  plugins = arg_get_value(options, "plugins");
  preferences = arg_get_value(options, "preferences");
  rules = arg_get_value(options, "rules");
  users = arg_get_value(options, "users");
#ifdef NESSUSNT
  /*
   * Just tell to the user that nessusd is running in background
   */
  create_thread(print_error, "nessusd is now running in background", 0);
  main_loop(sock, plugins, preferences, rules, users);
#else
  if(do_fork)
  {  	
   if(!fork()){
   	setsid();
   	main_loop(sock, plugins, preferences, rules, users);
	}
  }
  else main_loop(sock, plugins, preferences, rules, users);
#endif
  DO_EXIT(0);
  return(0);
}
 
 
