/*
 * QueSO
 *
 * By Jordi Murgo <savage@apostols.org>
 *
 *
 * Adapted as a Nessus plugin by Renaud Deraison <deraison@nessus.org>
 *
 */
 
#include <includes.h>


#define NAME "QueSO"
#define DESCRIPTION "This plugin attempts to guess the \n\
remote operating system type by sending \n\
specially forged TCP packets.\n\
If an attacker can guess your operating\n\
system, it will be able to break in more\n\
easily.\n\
Risk factor : Medium"

#define SUMMARY "Guesses the remote operating system"
#define COPYRIGHT "Jordi Murgo <savage@apostols.org>"


#define FILE_LOCATION "queso.conf location : "
#define FILE_DFL_LOC NESSUSD_DIR"queso.conf"
extern char * queso_main(struct arglist *, struct in_addr , int, char *, char *);


int
plugin_init(desc)
 struct arglist * desc;
{
 plug_set_name(desc, NAME);
 plug_set_description(desc, DESCRIPTION);
 plug_set_summary(desc, SUMMARY);
 plug_set_copyright(desc, COPYRIGHT);
 plug_set_family(desc, "Misc.");
 plug_set_category(desc, ACT_GATHER_INFO);
 add_plugin_preference(desc, FILE_LOCATION, PREF_ENTRY, FILE_DFL_LOC);
 return(0);
}

int
plugin_run(desc)
 struct arglist * desc;
{
 struct in_addr * host = plug_get_host_ip(desc);
 int port = plug_get_host_open_port(desc);
 char * filename = get_plugin_preference(desc, FILE_LOCATION);
 char * report;
 struct in_addr src;
 char * name = emalloc(255);
 char * inf;
 gethostname(name, 255);
 src = nn_resolve(name);
 efree(&name);
 inf = routethrough(host, &src);
 report = queso_main(desc, *host, port, filename, inf);
 if(report && !strstr(report, "Dead Host")){
   char* text = emalloc(strlen(report)+255);
   sprintf(text, "QueSO has found out that the remote host OS is \n%s\n", 
   		report);
   post_info(desc, -1, text);
   /*
    * Propagate the OS type among the plugins
    */
   if(strstr(report, "Wind"))
    {
    plug_set_key(desc, "OperatingSystem", ARG_STRING, "Windows");
    if(strstr(report, "Windoze NT")&&!strstr(report, "95"))
    plug_set_key(desc, "OperatingSystemVersion", ARG_STRING, "NT");
    }
   else if(strstr(report, "Linux")&&!strstr(report, "Standard"))
    plug_set_key(desc, "OperatingSystem", ARG_STRING, "Linux");
   else if(strstr(report, "Solaris"))
    plug_set_key(desc, "OperatingSystem", ARG_STRING, "Solaris");
   else plug_set_key(desc, "OperatingSystem", ARG_STRING,estrdup(report)); 
   efree(&text);
   efree(&report);
   }
 return(0);
}
 
