#include <includes.h>
#include <nessusraw.h>

struct pseudohdr
{
        struct in_addr saddr;
        struct in_addr daddr;
        u_char zero;
        u_char protocol;
        u_short length;
        struct tcphdr tcpheader;
};

static unsigned short tcp_in_cksum(unsigned short *ptr,int nbytes) {

register long           sum;            /* assumes long == 32 bits */
u_short                 oddbyte;
register u_short        answer;         /* assumes u_short == 16 bits */

/*
 * Our algorithm is simple, using a 32-bit accumulator (sum),
 * we add sequential 16-bit words to it, and at the end, fold back
 * all the carry bits from the top 16 bits into the lower 16 bits.
 */

sum = 0;
while (nbytes > 1)  {
sum += *ptr++;
nbytes -= 2;
}

/* mop up an odd byte, if necessary */
if (nbytes == 1) {
oddbyte = 0;            /* make sure top half is zero */
*((u_char *) &oddbyte) = *(u_char *)ptr;   /* one byte only */
sum += oddbyte;
}

/*
 * Add back carry outs from top 16 bits to low 16 bits.
 */

sum  = (sum >> 16) + (sum & 0xffff);    /* add high-16 to low-16 */
sum += (sum >> 16);                     /* add carry */
answer = ~sum;          /* ones-complement, then truncate to 16 bits */
return(answer);
}
unsigned char * 
forge_tcp_packet(source, dest, sport, dport, flags, seq)
 struct in_addr source;
 struct in_addr dest;
 unsigned int sport;
 unsigned int dport;
 int flags;
 int seq;
{
 u_char * ret = emalloc(sizeof(struct ip)+sizeof(struct tcphdr));
 struct pseudohdr pseudoheader;
 struct ip * ip = (struct ip *)ret;
 struct tcphdr * tcp = (struct tcphdr*)(ret + sizeof(struct ip));
 /*
  * Initialize the IP header
  */
 ip->ip_v = 4;
 ip->ip_hl = sizeof(struct ip)/4;
 ip->ip_id = random();
 ip->ip_tos = 0;
 ip->ip_len = FIX(sizeof(struct ip) + sizeof(struct tcphdr));
 ip->ip_ttl = 255;
 ip->ip_off = 0;
 ip->ip_p = IPPROTO_TCP;
 ip->ip_src.s_addr = source.s_addr;
 ip->ip_dst.s_addr = dest.s_addr;
 ip->ip_sum = 0;
 ip->ip_sum = tcp_in_cksum((short *)ip, sizeof(struct ip));
 
 tcp->th_sport = htons(sport);
 tcp->th_dport = htons(dport);
 tcp->th_seq = htonl(seq);
 tcp->th_ack = 0;
 tcp->th_flags = (flags);
#ifdef HAVE_TCPHDR_TH_OFF
 tcp->th_x2 = 0;
 tcp->th_off = 5;
#else 
 tcp->th_x2_off = 5 << 4;
#endif
 tcp->th_win = htons(4096);
 tcp->th_sum = 0;
 tcp->th_urp  = 0;
 
 bzero(&pseudoheader, 12+sizeof(struct tcphdr));
 pseudoheader.saddr.s_addr=source.s_addr;
 pseudoheader.daddr.s_addr=dest.s_addr;
 
 pseudoheader.protocol=6;
 pseudoheader.length=htons(sizeof(struct tcphdr));
 bcopy((char *) tcp,(char *) &pseudoheader.tcpheader,sizeof(struct tcphdr));
 tcp->th_sum = tcp_in_cksum((unsigned short *)&pseudoheader,12+sizeof(struct tcphdr));
 return(ret);
}
