#include <includes.h>
#include <pcap.h>
#include <nessusraw.h>
#include "../../nessusd/comm.h"

#define NAME "TCP NULL port scanning"
#define DESC "\n\
This scanner is an efficient way to perform\n\
a portscanning. It sends a tcp packet containing\n\
no flag set, and get the answer from the\n\
remote server.\n\
If the remote port is closed, then the remote host\n\
will send a packet containing the TH_RST flag set\n\
Because this check does not perform the three-way\n\
handshake, it may not be logged by older scan\n\
detectors.\n\
Beware : Windows boxes answer that every port is closed !"
#define SUMM "null scan"
#define COPY "no copyright"

extern unsigned char * 
forge_tcp_packet(struct in_addr, struct in_addr, unsigned int, unsigned int, int, int);


int 
plugin_init(desc)
 struct arglist* desc;
{
 plug_set_name(desc, NAME);
 plug_set_summary(desc, SUMM);
 plug_set_description(desc, DESC);
 plug_set_copyright(desc, COPY);
 plug_set_family(desc, "Port scanners");
 plug_set_category(desc, ACT_SCANNER);
 return(0);
}


pcap_t * init_pcap(src, dst, sport, dport)
 struct in_addr src;
 struct in_addr dst;
 int sport, dport;
{
 char errbuf[PCAP_ERRBUF_SIZE];
 pcap_t  * ret;
 char * interface;
 int promisc = 1;
 bpf_u_int32 netmask, network;
 struct bpf_program filter_prog;
 char * filter = emalloc(1024);
 char * a_src, *a_dst;
 
 a_src = estrdup(inet_ntoa(src));
 a_dst = estrdup(inet_ntoa(dst));
 sprintf(filter, "tcp and (src host %s and dst host %s and dst port %d)", 
 		  a_src,a_dst, dport);	
 efree(&a_src);
 efree(&a_dst);		 
 interface = routethrough(&src, &dst);
 if(!interface)interface = pcap_lookupdev(errbuf);
 if(pcap_lookupnet(interface, &network, &netmask, 0)<0)return(NULL);
 ret = pcap_open_live(interface, 92, 0, 40, errbuf);
 if(pcap_compile(ret, &filter_prog, filter, 1, netmask)<0)return(NULL);
 if(pcap_setfilter(ret, &filter_prog)<0)return(NULL);
 return(ret);
}



int decode_packet(
 unsigned short * ports,
 unsigned short * closed_ports,
 struct pcap_pkthdr * p,
 u_char * packet)
{
 struct ip * ip = (struct ip *)(packet);
 struct tcphdr * tcp; 
 int ip_opt;
 char * buffer = emalloc(1024);
 int ok = 0;
 
 
 bcopy(packet, buffer, p->len);
 ip_opt = ip->ip_hl;
 ip_opt = (ip_opt - 5)*4; 
 tcp = (struct tcphdr *)(packet + 20 + ip_opt);
 if(ip->ip_p == IPPROTO_TCP)ok=1;
 if(ok)
 {
 if((tcp->th_flags & TH_RST)){
  int i;
  int port = ntohs(tcp->th_sport);
  for(i=0;ports[i] && ports[i]!=port;i++);
  closed_ports[i] = 1; 
  }
  efree(&buffer);
  return(1);
 }
 efree(&buffer);
 return(0);
}

int
syn_scan(globals, dst, ports, hostinfos)
 struct arglist * globals;
 struct in_addr dst;
 unsigned short * ports;
 struct arglist * hostinfos;
{
 struct in_addr src;
 char * inf;
 char * my_name;
 int i = 0;
 int j = 0;
 int soc;
 int one = 1;
 struct sockaddr_in socaddr;
 ntp_caps * caps;
 pcap_t * pcap;
 int dlt_len;
 int end = 0;
 struct servent * service;
 unsigned short * closed_ports;
 char * pkt;
 int num_open = 0;
 
 soc = socket(AF_INET, SOCK_RAW, IPPROTO_RAW);
#ifdef IP_HDRINCL
 setsockopt(soc, IPPROTO_IP, IP_HDRINCL, &one, sizeof(one));
#endif
 if(soc < 0)return(0);
 
 caps = arg_get_value(globals, "ntp_caps");
 my_name = emalloc(255);
 gethostname(my_name, 255);
 src = nn_resolve(my_name);
 efree(&my_name);
 pcap = init_pcap(dst, src, 0, 123);
 if(!pcap)return(0);
 dlt_len = get_datalink_size(pcap_datalink(pcap));
 i = 0;
 while(ports[i++])end++;
 closed_ports = emalloc(sizeof(unsigned short)*(end+2));
 
 socaddr.sin_family = AF_INET;
 socaddr.sin_addr.s_addr = dst.s_addr;
 comm_send_status(globals, arg_get_value(hostinfos, "FQDN"),"portscan",  0,end);
 i = 0;
 while(ports[i])
 {
  int b;
  int ok = 0;
  struct pcap_pkthdr head;
  char * packet = forge_tcp_packet(src, dst, 123, ports[i],0, 1000+i);
  
  socaddr.sin_port = htons(ports[i]);
  if(sendto(soc, packet, sizeof(struct ip)+sizeof(struct tcphdr), 0,
       (struct sockaddr *)&socaddr, sizeof(struct sockaddr))<0);
  efree(&packet);
  packet = (void *)1;
  ok = 0;
  while(!ok && packet)
  {
   packet = (char *)pcap_next(pcap, &head);
   if(packet){
   	packet += dlt_len;
        ok = decode_packet(ports, closed_ports, &head, packet);
	}
  }
  i++;j++;
  if(j==100)
  {
     comm_send_status(globals, 
  		arg_get_value(hostinfos, "FQDN"),"portscan",  i,end);
	      j = 0;
  }
 }
 for(i=0;i<2;i++)
 do
  {
   struct pcap_pkthdr head;
   pkt = (char *)pcap_next(pcap, &head);
   if(pkt)
    {
    pkt += dlt_len;
    decode_packet(ports, closed_ports, &head, pkt);
    }
  }while(pkt);
 close(soc);
 comm_send_status(globals, arg_get_value(hostinfos, "FQDN"),"portscan",  end,end);
 pcap_close(pcap);
 for(i=0;ports[i];i++)if(!closed_ports[i])num_open++;
 if(num_open<200)for(i=0;ports[i];i++)
 {
  if(!closed_ports[i])
  {
  service = getservbyport(htons(ports[i]), "tcp");
  if(caps->ntp_11)
    auth_printf(globals,
	"SERVER <|> PORT <|> %s <|> %s (%d/tcp) <|> SERVER\n",
			  arg_get_value(hostinfos, "FQDN"),
                          service ? service->s_name : "unknown",
			  ports[i]);
   else
     auth_printf(globals,
	"SERVER <|> PORT <|> %s <|> %d <|> SERVER\n", 
			   arg_get_value(hostinfos, "FQDN"), 
			  ports[i]);
			   
  host_add_port(hostinfos, ports[i], 1, "");
  }
 }
 efree(&closed_ports);
 return(1);
}


int
plugin_run(env)
 struct arglist * env;
{
  short *ports;
  struct arglist * globals = arg_get_value(env, "globals");
  struct arglist * hostinfos = arg_get_value(env, "hostinfos");
  char * port_range = arg_get_value(env, "port_range");
  struct in_addr *p_addr;
  
  if(!strcmp("-1", port_range)||(!(ports = (short *)getpts(port_range))))return(0);
 
  if(!arg_get_value(hostinfos, "SCAN_RANGE"))
     arg_add_value(hostinfos, "SCAN_RANGE", ARG_PTR, sizeof(ports), ports);
  p_addr = arg_get_value(hostinfos, "IP");
  if(!p_addr)return(-1);

  if(syn_scan(globals, *p_addr, ports, hostinfos))
  {
    if(!arg_get_value(hostinfos, "scanned"))
    arg_add_value(hostinfos, "scanned", ARG_INT, sizeof(int), (void *)1);
  }
  return 0;
}

