/*
 * TCP Seq
 *
 */
 
#include <includes.h>
#include <pcap.h>
#include <nessusraw.h>


#define NAME "TCP Sequence Prediction"
#define DESC "This plugin checks whether it is\n\
easy for an attacker or not to establish a spoofed\n\
TCP connection to the remote host, by examining\n\
the TCP sequence numbers issued by the remote host.\n\
If this plugin finds a problem, we recommand\n\
that you upgrade your system, contact your vendor\n\
for a patch or change operating system.\n\
Risk factor : High"
#define COPY "distributed under the GPL"
#define SUMM "tcp.th_seq check"

struct pseudohdr
{
        struct in_addr saddr;
        struct in_addr daddr;
        u_char zero;
        u_char protocol;
        u_short length;
        struct tcphdr tcpheader;
};

struct cb_infos 
{
 	unsigned long seq1;
	unsigned long seq2;
	unsigned int dl_len;
};

#define SEQ_CONSTANT 0
#define SEQ_64K      1
#define SEQ_i800     2
#define SEQ_UNKNOWN  3

unsigned short tcp_in_cksum(unsigned short *ptr,int nbytes) {

register long           sum;            /* assumes long == 32 bits */
unsigned short                 oddbyte;
register unsigned short        answer;         /* assumes u_short == 16 bits */

/*
 * Our algorithm is simple, using a 32-bit accumulator (sum),
 * we add sequential 16-bit words to it, and at the end, fold back
 * all the carry bits from the top 16 bits into the lower 16 bits.
 */

sum = 0;
while (nbytes > 1)  {
sum += *ptr++;
nbytes -= 2;
}

/* mop up an odd byte, if necessary */
if (nbytes == 1) {
oddbyte = 0;            /* make sure top half is zero */
*((u_char *) &oddbyte) = *(u_char *)ptr;   /* one byte only */
sum += oddbyte;
}

/*
 * Add back carry outs from top 16 bits to low 16 bits.
 */

sum  = (sum >> 16) + (sum & 0xffff);    /* add high-16 to low-16 */
sum += (sum >> 16);                     /* add carry */
answer = ~sum;          /* ones-complement, then truncate to 16 bits */
return(answer);
}

unsigned char * 
forge_tcp_packet(source, dest, sport, dport, flags, seq)
 struct in_addr source;
 struct in_addr dest;
 unsigned int sport;
 unsigned int dport;
 int flags;
 int seq;
{
 u_char * ret = emalloc(sizeof(struct ip)+sizeof(struct tcphdr));
 struct pseudohdr pseudoheader;
 struct ip * ip = (struct ip *)ret;
 struct tcphdr * tcp = (struct tcphdr*)(ret + sizeof(struct ip));
 /*
  * Initialize the IP header
  */
 ip->ip_v = 4;
 ip->ip_hl = sizeof(struct ip)/4;
 ip->ip_id = random();
 ip->ip_tos = 0;
 ip->ip_len = FIX(sizeof(struct ip) + sizeof(struct tcphdr));
 ip->ip_ttl = 255;
 ip->ip_off = 0;
 ip->ip_p = IPPROTO_TCP;
 ip->ip_src.s_addr = source.s_addr;
 ip->ip_dst.s_addr = dest.s_addr;
 ip->ip_sum = 0;
 ip->ip_sum = tcp_in_cksum((short *)ip, sizeof(struct ip));
 
 tcp->th_sport = htons(sport);
 tcp->th_dport = htons(dport);
 tcp->th_seq = htonl(seq);
 tcp->th_ack = 0;
 tcp->th_flags = flags;
#ifndef HAVE_TCPHDR_TH_X2_OFF
 tcp->th_x2 = 0;
 tcp->th_off = 5;
#else
 tcp->th_x2_off = 5 << 4;
#endif

 tcp->th_win = htons(4096);
 tcp->th_sum = 0;
 tcp->th_urp  = 0;
 
 bzero(&pseudoheader, 12+sizeof(struct tcphdr));
 pseudoheader.saddr.s_addr=source.s_addr;
 pseudoheader.daddr.s_addr=dest.s_addr;
 
 pseudoheader.protocol=6;
 pseudoheader.length=htons(sizeof(struct tcphdr));
 bcopy((char *) tcp,(char *) &pseudoheader.tcpheader,sizeof(struct tcphdr));
 tcp->th_sum = tcp_in_cksum((unsigned short *)&pseudoheader,12+sizeof(struct tcphdr));
 return(ret);
}

unsigned int
euclid_gcd(a,b)
 unsigned int a;
 unsigned int b;
{
 if(a < b)return(euclid_gcd(b,a));
 if(!b)return(a);
 return(euclid_gcd(b,a%b));
}


int 
analyze_seq(unsigned int Seq[10], unsigned int Conf[10])
{
 unsigned int diffs[9];
 int i, n = 0;
 unsigned int gcd;
 unsigned int seq[10];
 
 for(i=0;i<10;i++)seq[i]=0;
 for(i=0;i<10;i++)if(Conf[i])seq[n++]=Seq[i];
 if(!n)return(SEQ_UNKNOWN);				
 for(i=0;i<9;i++)diffs[i] = 0;
 for(i=0;i<n-1;i++){
    diffs[i]=seq[i+1]-seq[i];
   }

 gcd = diffs[0];
 for(i=1;i<n;i++)gcd = euclid_gcd(gcd, diffs[i]);
 if(!gcd)return(SEQ_CONSTANT);
 else if((gcd % 64000)==0)return(SEQ_64K);
 else if((gcd % 800)==0)return(SEQ_i800);
 return(SEQ_UNKNOWN);
}

int
plugin_init(desc)
 struct arglist * desc;
{
 plug_set_name(desc, NAME);
 plug_set_description(desc, DESC);
 plug_set_summary(desc, SUMM);
 plug_set_copyright(desc, COPY);
 plug_set_category(desc, ACT_GATHER_INFO);
 plug_set_family(desc, "Misc.");
 return(0);
}

pcap_t * init_pcap(src, dst, sport, dport)
 struct in_addr src;
 struct in_addr dst;
 int sport, dport;
{
 char errbuf[PCAP_ERRBUF_SIZE];
 pcap_t  * ret;
 char * interface;
 bpf_u_int32 netmask, network;
 struct bpf_program filter_prog;
 char * filter = emalloc(1024);
 char * a_src, *a_dst;
 
 a_src = estrdup(inet_ntoa(src));
 a_dst = estrdup(inet_ntoa(dst));
 sprintf(filter, "tcp and (src host %s and src port %d and dst host %s \
and dst port %d)", 
 		  a_src,sport,a_dst, dport);	 		  
 interface = routethrough(&src, &dst);
 
 if(!interface)interface = pcap_lookupdev(errbuf);
 if(pcap_lookupnet(interface, &network, &netmask, 0)<0)return(NULL);
 ret = pcap_open_live(interface, 92, 0, 180, errbuf);
 if(!ret)return(NULL);
 if(pcap_compile(ret, &filter_prog, filter, 1, netmask)<0){
        return(NULL);
     }
 if(pcap_setfilter(ret, &filter_prog)<0){
	return(NULL);
 }
 efree(&filter);
 return(ret);
}


int seq_decode_packet(
 struct cb_infos * infos,
 struct pcap_pkthdr * p,
 u_char * packet)
{
 struct tcphdr * tcp = (struct tcphdr *)(packet+20+infos->dl_len);
 if(!packet)return(0);
 infos->seq1 = ntohl(tcp->th_seq);
 infos->seq2 = ntohl(tcp->th_ack) - 1000;
 return(0);
}


int
plugin_run(desc)
 struct arglist * desc;
{
 pcap_t * pcap;
 int soc;
 struct sockaddr_in saddr;
 struct in_addr source, dest;
 struct in_addr *t = plug_get_host_ip(desc);
 int sport, dport;
 int one = 1;
 char * name = emalloc(255);
 int i = 0; 
 struct pcap_pkthdr head;
 char* p;
 unsigned int seq[10][2];
 unsigned int Seq[10];
 unsigned int Conf[10];
 int j;
 int dlt_len;
 for(i=0;i<2;i++)
  for(j=0;j<10;j++)seq[j][i] = 0;
  
 soc = socket(AF_INET, SOCK_RAW, IPPROTO_RAW);
#ifdef IP_HDRINCL
 setsockopt(soc, IPPROTO_IP, IP_HDRINCL, &one, sizeof(one));
#endif
 gethostname(name, 254);
 source = nn_resolve(name);
 dest.s_addr   = t->s_addr;
 sport = 123;
 dport = plug_get_host_open_port(desc);
 if(!dport)return(0);
 pcap = init_pcap(dest, source, dport, sport);
 if(!pcap)return(0);
 saddr.sin_family = AF_INET;
 saddr.sin_port = htons(dport);
 saddr.sin_addr.s_addr = dest.s_addr;
 dlt_len = get_datalink_size(pcap_datalink(pcap));
 for(i=0;i<10;i++)
 {
  struct cb_infos * infos = emalloc(sizeof(struct cb_infos));
  u_char * packet = forge_tcp_packet(source, dest, sport, dport, TH_SYN,1000+i);
  if((sendto(soc, packet, sizeof(struct ip)+sizeof(struct tcphdr), 0,
        (struct sockaddr *)&saddr, sizeof(struct sockaddr_in)))<0)
	perror("sendto ");
  free(packet);
  infos->dl_len = dlt_len;
  pcap_dispatch(pcap, 1, (pcap_handler)seq_decode_packet, (void *)infos);
  seq[i][0] = infos->seq1;
  seq[i][1] = infos->seq2;
  efree(&infos);
 }
 
 for(i=0;i<10;i++)Seq[i] = seq[i][0];
 for(i=0;i<10;i++)Conf[i] = seq[i][1];
 switch(analyze_seq(Seq, Conf))
 {
  case SEQ_CONSTANT :
   post_hole(desc, -1, "\
The remote host can easily be spoofed :\n\
its TCP sequence number is constant !\n\
Solution : contact your vendor for a fix\n\
or change operating systems");
   break;
 case SEQ_64K :
  post_hole(desc, -1, "\
The remote host can easily be spoofed :\n\
its TCP sequence prediction uses the old\n\
64K rule (each TCP sequence number is\n\
incremented by 64000).\n\
Solution : contact your vendor for a fix\n\
or change operating systems");
  break;
case SEQ_i800 :
post_hole(desc, -1, "\
The remote host can easily be spoofed :\n\
its TCP sequence prediction uses the\n\
800 rule (each TCP sequence number is\n\
incremented by 800).\n\
Solution : contact your vendor for a fix\n\
or change operating systems");
 break;
 } 
 pcap_close(pcap);
 close(soc);
 return(0);
}	
 	
